############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from enum import Enum, unique
from .common import junctions_from_blocks
from .id_policy import SimpleIDDistributor
from .serialization import *
from .polya_finder import PolyAInfo

logger = logging.getLogger('IsoQuant')


@unique
class ReadAssignmentType(Enum):
    unique = 1
    noninformative = 0
    intergenic = 20
    ambiguous = 10
    unique_minor_difference = 2
    inconsistent = 30
    inconsistent_non_intronic = 31
    inconsistent_ambiguous = 32
    suspended = 255

    def is_inconsistent(self):
        return self in [ReadAssignmentType.inconsistent,
                        ReadAssignmentType.inconsistent_ambiguous,
                        ReadAssignmentType.inconsistent_non_intronic]

    def is_consistent(self):
        return self in [ReadAssignmentType.unique,
                        ReadAssignmentType.unique_minor_difference,
                        ReadAssignmentType.ambiguous]

    def is_unassigned(self):
        return self in [ReadAssignmentType.noninformative,
                        ReadAssignmentType.intergenic]

    def is_unique(self):
        return self in [ReadAssignmentType.unique_minor_difference,
                        ReadAssignmentType.unique]

    def is_ambiguous(self):
        return self in [ReadAssignmentType.ambiguous,
                        ReadAssignmentType.inconsistent_ambiguous]

# SQANTI-like
@unique
class MatchClassification(Enum):
    undefined = 0
    full_splice_match = 10
    incomplete_splice_match = 11
    mono_exon_match = 12
    novel_in_catalog = 20
    novel_not_in_catalog = 21
    genic = 30
    antisense = 40
    fusion = 50
    intergenic = 32
    genic_intron = 31

    @staticmethod
    def get_inconsistency_classification(match_event_subtypes):
        if any(me.event_type in nnic_event_types for me in match_event_subtypes):
            return MatchClassification.novel_not_in_catalog
        elif any(me.event_type in nic_event_types for me in match_event_subtypes):
            return MatchClassification.novel_in_catalog
        return MatchClassification.undefined

    @staticmethod
    def get_mono_exon_classification(match_event_subtypes):
        # events are not mixed in the list
        if any(e.event_type in {MatchEventSubtype.alternative_polya_site_left, MatchEventSubtype.alternative_polya_site_right,
                                MatchEventSubtype.internal_polya_left, MatchEventSubtype.internal_polya_right}
               for e in match_event_subtypes):
            return MatchClassification.novel_not_in_catalog
        if any(e.event_type == MatchEventSubtype.unspliced_intron_retention for e in match_event_subtypes):
            return MatchClassification.novel_in_catalog
        elif any(e.event_type in {MatchEventSubtype.incomplete_intron_retention_left,
                                  MatchEventSubtype.incomplete_intron_retention_right} for e in match_event_subtypes):
            return MatchClassification.genic
        elif match_event_subtypes[0].event_type == MatchEventSubtype.fake_micro_intron_retention:
            return MatchClassification.incomplete_splice_match
        elif match_event_subtypes[0].event_type == MatchEventSubtype.mono_exon_match:
            return MatchClassification.mono_exon_match
        elif match_event_subtypes[0].event_type == MatchEventSubtype.mono_exonic:
            return MatchClassification.incomplete_splice_match
        else:
            logger.warning("Unexpected set of monoexonic events: " + str(match_event_subtypes))
            return MatchClassification.undefined


@unique
class MatchEventSubtype(Enum):
    none = 0
    undefined = 1
    # non-contradictory
    mono_exonic = 10
    antisense = 9
    fsm = 11
    ism_left = 15
    ism_right = 13
    ism_internal = 14
    mono_exon_match = 19
    # alignment artifacts
    intron_shift = 21
    exon_misalignment = 22
    fake_terminal_exon_left = 24
    fake_terminal_exon_right = 26
    terminal_exon_misalignment_right = 28
    terminal_exon_misalignment_left = 29
    # minor alternations
    exon_elongation_left = 25
    exon_elongation_right = 23
    # intron retentions
    intron_retention = 31
    unspliced_intron_retention = 32
    incomplete_intron_retention_left = 38
    incomplete_intron_retention_right = 39
    fake_micro_intron_retention = 34
    # major alternation
    # alternative donor/acceptor sites
    alt_left_site_known = 101
    alt_right_site_known = 102
    alt_left_site_novel = 103
    alt_right_site_novel = 104
    # additional introns in the middle
    extra_intron_novel = 1012
    extra_intron_known = 1011
    # extra inrons on the sides
    extra_intron_flanking_left = 1013
    extra_intron_flanking_right = 1015
    # significant exon elongation, more than allowed
    major_exon_elongation_left = 1005
    major_exon_elongation_right = 1003
    # other intron modifications
    intron_migration = 114
    intron_alternation_novel = 115
    intron_alternation_known = 116
    # mutually exclusive
    mutually_exclusive_exons_novel = 121
    mutually_exclusive_exons_known = 122
    # exon skipping
    exon_skipping_known = 133
    exon_skipping_novel = 134
    # similar to mutually exclusive exons but when read exon is attached to another exon
    exon_merge_known = 135
    exon_merge_novel = 136
    # exon gain
    exon_gain_known = 145
    exon_gain_novel = 146
    # similar to mutually exclusive exons but when reference exon is attached to another exon
    exon_detach_known = 147
    exon_detach_novel = 148
    # terminal exon shift
    terminal_exon_shift_known = 151
    terminal_exon_shift_novel = 152
    # other
    alternative_structure_novel = 181
    alternative_structure_known = 182
    # TTS and TSS
    alternative_polya_site_right = 200
    alternative_polya_site_left = 201
    internal_polya_left = 202
    internal_polya_right = 203
    alternative_tss_left = 204
    alternative_tss_right = 205
    correct_polya_site_left = 222
    correct_polya_site_right = 223
    aligned_polya_tail = 300

    terminal_site_match_left = 250
    terminal_site_match_right = 251
    terminal_site_match_left_precise = 252
    terminal_site_match_right_precise = 253

    def __lt__(self, other):
        return self.value < other.value

    @staticmethod
    def is_alignment_artifact(match_event_subtype):
        return match_event_subtype in {MatchEventSubtype.intron_shift, MatchEventSubtype.exon_misalignment,
                                       MatchEventSubtype.terminal_exon_misalignment_right,
                                       MatchEventSubtype.terminal_exon_misalignment_left,
                                       MatchEventSubtype.fake_terminal_exon_left,
                                       MatchEventSubtype.fake_terminal_exon_right,
                                       MatchEventSubtype.fake_micro_intron_retention}

    @staticmethod
    def is_minor_error(match_event_subtype):
        return match_event_subtype in {MatchEventSubtype.exon_elongation_left,
                                       MatchEventSubtype.exon_elongation_right,
                                       MatchEventSubtype.intron_shift,
                                       MatchEventSubtype.exon_misalignment,
                                       MatchEventSubtype.fake_micro_intron_retention,
                                       MatchEventSubtype.fake_terminal_exon_left,
                                       MatchEventSubtype.fake_terminal_exon_right,
                                       MatchEventSubtype.terminal_exon_misalignment_left,
                                       MatchEventSubtype.terminal_exon_misalignment_right}

    @staticmethod
    def is_consistent(match_event_subtype):
        return match_event_subtype in {MatchEventSubtype.none,
                                       MatchEventSubtype.mono_exonic,
                                       MatchEventSubtype.mono_exon_match,
                                       MatchEventSubtype.fsm,
                                       MatchEventSubtype.ism_left,
                                       MatchEventSubtype.ism_right,
                                       MatchEventSubtype.ism_internal,
                                       MatchEventSubtype.correct_polya_site_left,
                                       MatchEventSubtype.correct_polya_site_right,
                                       MatchEventSubtype.terminal_site_match_left,
                                       MatchEventSubtype.terminal_site_match_right,
                                       MatchEventSubtype.terminal_site_match_left_precise,
                                       MatchEventSubtype.terminal_site_match_right_precise}

    @staticmethod
    def is_major_elongation(match_event_subtype):
        return match_event_subtype in {MatchEventSubtype.major_exon_elongation_left,
                                       MatchEventSubtype.major_exon_elongation_right}

    @staticmethod
    def is_minor_elongation(match_event_subtype):
        return match_event_subtype in {MatchEventSubtype.exon_elongation_left,
                                       MatchEventSubtype.exon_elongation_right}

    @staticmethod
    def is_major_inconsistency(match_event_subtype):
        return match_event_subtype in all_major_events

    @staticmethod
    def is_intronic_inconsistency(match_event_subtype):
        return match_event_subtype in intronic_major_events


event_subtype_cost = {
    MatchEventSubtype.none:0,
    MatchEventSubtype.undefined:0,
    MatchEventSubtype.mono_exonic:0,
    MatchEventSubtype.fsm:0,
    MatchEventSubtype.ism_left:0,
    MatchEventSubtype.ism_right:0,
    MatchEventSubtype.ism_internal:0,
    MatchEventSubtype.mono_exon_match:0,
    MatchEventSubtype.intron_shift:0.1,
    MatchEventSubtype.exon_misalignment:0.1,
    MatchEventSubtype.terminal_exon_misalignment_right:0.1,
    MatchEventSubtype.terminal_exon_misalignment_left:0.1,
    MatchEventSubtype.fake_terminal_exon_left:0.2,
    MatchEventSubtype.fake_terminal_exon_right:0.2,
    # minor alternations
    MatchEventSubtype.exon_elongation_left:0.1,
    MatchEventSubtype.exon_elongation_right:0.1,
    # intron retentions
    MatchEventSubtype.intron_retention:0.6,
    MatchEventSubtype.unspliced_intron_retention:0.6,
    MatchEventSubtype.incomplete_intron_retention_left:0.7,
    MatchEventSubtype.incomplete_intron_retention_right:0.7,
    MatchEventSubtype.fake_micro_intron_retention:0.1,
    # major alternation
    # alternative donor/acceptor sites
    MatchEventSubtype.alt_left_site_known:1,
    MatchEventSubtype.alt_right_site_known:1,
    MatchEventSubtype.alt_left_site_novel:1,
    MatchEventSubtype.alt_right_site_novel:1,
    # additional introns in the middle
    MatchEventSubtype.extra_intron_novel:1,
    MatchEventSubtype.extra_intron_known:1,
    # extra inrons on the sides
    MatchEventSubtype.extra_intron_flanking_left:1,
    MatchEventSubtype.extra_intron_flanking_right:1,
    # significant exon elongation, more than allowed
    MatchEventSubtype.major_exon_elongation_left:0.6,
    MatchEventSubtype.major_exon_elongation_right:0.6,
    # other intron modifications
    MatchEventSubtype.intron_migration:1,
    MatchEventSubtype.intron_alternation_novel:1,
    MatchEventSubtype.intron_alternation_known:1,
    # mutually exclusive
    MatchEventSubtype.mutually_exclusive_exons_novel:0.8,
    MatchEventSubtype.mutually_exclusive_exons_known:0.8,
    # exon skipping
    MatchEventSubtype.exon_skipping_known:1,
    MatchEventSubtype.exon_skipping_novel:1,
    MatchEventSubtype.exon_merge_known:0.75,
    MatchEventSubtype.exon_merge_novel:0.75,
    # exon gain
    MatchEventSubtype.exon_gain_known:1,
    MatchEventSubtype.exon_gain_novel:1,
    MatchEventSubtype.exon_detach_known:0.75,
    MatchEventSubtype.exon_detach_novel:0.75,
    MatchEventSubtype.terminal_exon_shift_known:0.5,
    MatchEventSubtype.terminal_exon_shift_novel:0.5,
    # other
    MatchEventSubtype.alternative_structure_novel:1,
    MatchEventSubtype.alternative_structure_known:1,
    # TTS and TSS
    MatchEventSubtype.alternative_polya_site_left:0.6,
    MatchEventSubtype.alternative_polya_site_right:0.6,
    MatchEventSubtype.internal_polya_left:0.5,
    MatchEventSubtype.internal_polya_right:0.5,
    MatchEventSubtype.alternative_tss_left :0.6,
    MatchEventSubtype.alternative_tss_right :0.6,
    MatchEventSubtype.correct_polya_site_left:0,
    MatchEventSubtype.correct_polya_site_right:0,
    MatchEventSubtype.aligned_polya_tail:0,
    MatchEventSubtype.terminal_site_match_left:0,
    MatchEventSubtype.terminal_site_match_right:0,
    MatchEventSubtype.terminal_site_match_left_precise:0,
    MatchEventSubtype.terminal_site_match_right_precise:0
}


def elongation_cost(params, elongation_len):
    min_cost = event_subtype_cost[MatchEventSubtype.exon_elongation_left]
    max_cost = event_subtype_cost[MatchEventSubtype.major_exon_elongation_left]
    lower_bound = params.minor_exon_extension
    upper_bound = params.major_exon_extension
    if elongation_len <= lower_bound:
        return min_cost
    elif elongation_len >= upper_bound:
        return max_cost
    else:
        return min_cost + (max_cost - min_cost) * (elongation_len - lower_bound) / (upper_bound - lower_bound)


nnic_event_types = {
    MatchEventSubtype.alt_left_site_novel, MatchEventSubtype.alt_right_site_novel,
    MatchEventSubtype.extra_intron_novel, MatchEventSubtype.extra_intron_flanking_left,
    MatchEventSubtype.extra_intron_flanking_right, MatchEventSubtype.mutually_exclusive_exons_novel,
    MatchEventSubtype.exon_gain_novel, MatchEventSubtype.exon_skipping_novel,
    MatchEventSubtype.exon_detach_novel, MatchEventSubtype.exon_merge_novel,
    MatchEventSubtype.terminal_exon_shift_novel,
    MatchEventSubtype.alternative_structure_novel, MatchEventSubtype.intron_alternation_novel,
    MatchEventSubtype.alternative_polya_site_left, MatchEventSubtype.alternative_polya_site_right,
    MatchEventSubtype.alternative_tss_right, MatchEventSubtype.alternative_tss_left
}

nic_event_types = {
    MatchEventSubtype.unspliced_intron_retention, MatchEventSubtype.intron_retention,
    MatchEventSubtype.alt_left_site_known, MatchEventSubtype.alt_right_site_known,
    MatchEventSubtype.extra_intron_known, MatchEventSubtype.intron_migration,
    MatchEventSubtype.mutually_exclusive_exons_known, MatchEventSubtype.exon_skipping_known,
    MatchEventSubtype.exon_detach_known, MatchEventSubtype.exon_merge_known,
    MatchEventSubtype.terminal_exon_shift_known,
    MatchEventSubtype.exon_gain_known, MatchEventSubtype.alternative_structure_known,
    MatchEventSubtype.intron_alternation_known, MatchEventSubtype.major_exon_elongation_left,
    MatchEventSubtype.major_exon_elongation_right, MatchEventSubtype.incomplete_intron_retention_left,
    MatchEventSubtype.incomplete_intron_retention_right, MatchEventSubtype.internal_polya_right,
    MatchEventSubtype.internal_polya_left
}

nonintronic_events = {
    MatchEventSubtype.alternative_polya_site_left, MatchEventSubtype.alternative_polya_site_right,
    MatchEventSubtype.alternative_tss_left, MatchEventSubtype.alternative_tss_right,
    MatchEventSubtype.internal_polya_right, MatchEventSubtype.internal_polya_left,
    MatchEventSubtype.major_exon_elongation_left, MatchEventSubtype.major_exon_elongation_right,
    MatchEventSubtype.exon_elongation_left, MatchEventSubtype.exon_elongation_right,
}

all_major_events = nic_event_types.union(nnic_event_types)

intronic_major_events = all_major_events.difference(nonintronic_events)


# (side, is_known) -> alternation type
alternative_sites = {("left", True): MatchEventSubtype.alt_left_site_known,
                     ("left", False): MatchEventSubtype.alt_left_site_novel,
                     ("right", True): MatchEventSubtype.alt_right_site_known,
                     ("right", False): MatchEventSubtype.alt_right_site_novel}


class SupplementaryMatchConstants:
    extra_left_mod_position = (1 << 30) - 1
    extra_right_mod_position = (1 << 30) + 1
    undefined_position = (1 << 31)
    undefined_region = (undefined_position, undefined_position)
    extra_left_region = (extra_left_mod_position, extra_left_mod_position)
    extra_right_region = (extra_right_mod_position, extra_right_mod_position)
    absent_position = (1 << 31) - 1


class MatchEvent:
    def __init__(self, event_type:MatchEventSubtype,
                 isoform_region:tuple=SupplementaryMatchConstants.undefined_region,
                 read_region:tuple=SupplementaryMatchConstants.undefined_region,
                 event_info=0):
        self.event_type = event_type
        self.isoform_region = isoform_region
        self.read_region = read_region
        self.event_info = event_info

    @classmethod
    def deserialize(cls, infile):
        event = cls.__new__(cls)
        event.event_type = MatchEventSubtype(read_int(infile, SHORT_INT_BYTES))
        # TODO: think about 2/3-byte storage
        event.isoform_region = (read_int(infile), read_int(infile))
        event.read_region = (read_int(infile), read_int(infile))
        event.event_info = read_int_neg(infile)
        return event

    def serialize(self, outfile):
        write_int(self.event_type.value, outfile, SHORT_INT_BYTES)
        write_int(self.isoform_region[0], outfile)
        write_int(self.isoform_region[1], outfile)
        write_int(self.read_region[0], outfile)
        write_int(self.read_region[1], outfile)
        write_int_neg(self.event_info, outfile)

    def __repr__(self):
        return "%s:%s,%s,%s" % (self.event_type.name, str(self.isoform_region),
                                str(self.read_region), str(self.event_info))


class IsoformMatch:
    """
    Represents a match between a read and an isoform.

    Memory Optimization:
        Gene and transcript IDs are stored as integers referencing
        shared string pools (string_pools parameter required).
    """
    def __init__(self, match_classification, string_pools, assigned_gene=None, assigned_transcript=None,
                 match_subclassification = None, transcript_strand='.', penalty_score=0):
        # Store string pools reference (required for memory optimization)
        assert string_pools is not None, "string_pools is required"
        self._string_pools = string_pools

        # Store gene/transcript as integers
        self.assigned_gene_id = string_pools.gene_pool.get_int(assigned_gene) if assigned_gene else None
        self.assigned_transcript_id = string_pools.transcript_pool.get_int(assigned_transcript) if assigned_transcript else None

        self.transcript_strand = transcript_strand  # Keep as char (1 byte)
        self.match_classification = match_classification
        self.penalty_score = penalty_score
        if match_subclassification is None:
            self.match_subclassifications = []
        elif isinstance(match_subclassification, list):
            self.match_subclassifications = \
                list(filter(lambda x: x.event_type != MatchEventSubtype.none, match_subclassification))
        else:
            self.match_subclassifications = [match_subclassification]

    @property
    def assigned_gene(self):
        """Return gene string (backward compatibility)"""
        return self._string_pools.gene_pool.get_str(self.assigned_gene_id) if self.assigned_gene_id is not None else None

    @assigned_gene.setter
    def assigned_gene(self, value):
        """Set gene from string"""
        self.assigned_gene_id = self._string_pools.gene_pool.get_int(value) if value is not None else None

    @property
    def assigned_transcript(self):
        """Return transcript string (backward compatibility)"""
        return self._string_pools.transcript_pool.get_str(self.assigned_transcript_id) if self.assigned_transcript_id is not None else None

    @assigned_transcript.setter
    def assigned_transcript(self, value):
        """Set transcript from string"""
        self.assigned_transcript_id = self._string_pools.transcript_pool.get_int(value) if value is not None else None

    @classmethod
    def deserialize(cls, infile, string_pools):
        assert string_pools is not None, "string_pools is required"
        match = cls.__new__(cls)
        match._string_pools = string_pools
        match.assigned_gene_id = read_int_or_none(infile)
        match.assigned_transcript_id = read_int_or_none(infile)
        match.transcript_strand = read_string(infile)
        match.match_classification = MatchClassification(read_short_int(infile))
        match.penalty_score = float(read_int(infile)) / float(SHORT_FLOAT_MULTIPLIER)
        match.match_subclassifications = read_list(infile, MatchEvent.deserialize)
        return match

    def serialize(self, outfile):
        write_int_or_none(self.assigned_gene_id, outfile)
        write_int_or_none(self.assigned_transcript_id, outfile)
        write_string(self.transcript_strand, outfile)
        write_short_int(self.match_classification.value, outfile)
        write_int(int(self.penalty_score * SHORT_FLOAT_MULTIPLIER), outfile)
        write_list(self.match_subclassifications, outfile, MatchEvent.serialize)

    def add_subclassification(self, match_subclassification):
        if len(self.match_subclassifications) == 1 and \
                self.match_subclassifications[0].event_type in {MatchEventSubtype.undefined, MatchEventSubtype.none}:
            self.match_subclassifications = [match_subclassification]
        else:
            self.match_subclassifications.append(match_subclassification)

    def set_classification(self, classification):
        self.match_classification = classification

    def monoexon_is_consistent(self):
        valid_subtypes = {MatchEventSubtype.none, MatchEventSubtype.mono_exonic, MatchEventSubtype.mono_exon_match}
        return all(el.event_type in valid_subtypes for el in self.match_subclassifications)


class BasicReadAssignment:
    """
    Simplified read assignment for serialization and multimap resolution.

    Memory Optimization:
        chr_id, barcode, umi, gene IDs, and isoform IDs are stored as integers
        referencing shared string pools (string_pools parameter required).
    """
    def __init__(self, read_assignment):
        self.assignment_id = read_assignment.assignment_id
        self.read_id = read_assignment.read_id

        # String interning for memory optimization
        assert hasattr(read_assignment, '_string_pools'), "read_assignment must have string_pools"
        self._string_pools = read_assignment._string_pools
        assert self._string_pools is not None, "string_pools is required"

        # Copy interned fields
        self.chr_id_int = read_assignment.chr_id_int
        self.barcode_id = read_assignment.barcode_id
        self.umi_id = read_assignment.umi_id

        self.start = 0
        self.end = 0
        if read_assignment.exons:
            self.start = read_assignment.exons[0][0]
            self.end = read_assignment.exons[-1][1]
        self.genomic_region = read_assignment.genomic_region
        self.multimapper = read_assignment.multimapper
        self.polyA_found = read_assignment.polyA_found
        self.assignment_type = read_assignment.assignment_type
        self.gene_assignment_type = read_assignment.gene_assignment_type
        self.penalty_score = 0.0

        # Store gene/isoform lists as integer IDs
        self.gene_ids = []
        self.isoform_ids = []

        if read_assignment.isoform_matches:
            gene_set = set()
            isoform_set = set()
            for m in read_assignment.isoform_matches:
                self.penalty_score = min(self.penalty_score, read_assignment.isoform_matches[0].penalty_score)
                if m.assigned_gene:
                    gene_set.add(m.assigned_gene_id)
                if m.assigned_transcript:
                    isoform_set.add(m.assigned_transcript_id)

            self.gene_ids = list(gene_set)
            self.isoform_ids = list(isoform_set)

    @property
    def chr_id(self):
        """Return chromosome string"""
        if self.chr_id_int is not None:
            return self._string_pools.chromosome_pool.get_str(self.chr_id_int)
        # Handle unpickled state where chr_id is stored as string
        if hasattr(self, '_chr_id_str'):
            return self._chr_id_str
        return "."

    @chr_id.setter
    def chr_id(self, value):
        """Set chromosome from string"""
        self.chr_id_int = self._string_pools.chromosome_pool.get_int(value) if value else None

    @property
    def barcode(self):
        """Return barcode string"""
        if self.barcode_id is not None:
            return self._string_pools.barcode_pool.get_str(self.barcode_id)
        return None

    @barcode.setter
    def barcode(self, value):
        """Set barcode from string"""
        self.barcode_id = self._string_pools.barcode_pool.get_int(value) if value else None

    @property
    def umi(self):
        """Return UMI string"""
        if self.umi_id is not None:
            return self._string_pools.umi_pool.get_str(self.umi_id)
        return None

    @umi.setter
    def umi(self, value):
        """Set UMI from string"""
        self.umi_id = self._string_pools.umi_pool.get_int(value) if value else None

    @property
    def genes(self):
        """Return list of gene strings"""
        if self.gene_ids:
            return [self._string_pools.gene_pool.get_str(gid) for gid in self.gene_ids]
        # Handle unpickled state where genes are stored as strings
        if hasattr(self, '_genes'):
            return self._genes
        return []

    @genes.setter
    def genes(self, value):
        """Set genes from list of strings"""
        self.gene_ids = [self._string_pools.gene_pool.get_int(g) for g in value]

    @property
    def isoforms(self):
        """Return list of isoform strings"""
        if self.isoform_ids:
            return [self._string_pools.transcript_pool.get_str(tid) for tid in self.isoform_ids]
        # Handle unpickled state where isoforms are stored as strings
        if hasattr(self, '_isoforms'):
            return self._isoforms
        return []

    @isoforms.setter
    def isoforms(self, value):
        """Set isoforms from list of strings"""
        self.isoform_ids = [self._string_pools.transcript_pool.get_int(t) for t in value]

    def __eq__(self, other):
        if isinstance(other, BasicReadAssignment):
            return (self.read_id == other.read_id and
                    self.chr_id == other.chr_id and
                    self.start == other.start and
                    self.end == other.end and
                    self.isoforms == other.isoforms)
        return False

    def __getstate__(self):
        # When pickling, always use string representation for compatibility
        return (self.assignment_id,
                self.read_id,
                self.chr_id,  # Property will convert from int if needed
                self.start,
                self.end,
                self.genomic_region[0],
                self.genomic_region[1],
                self.multimapper,
                self.polyA_found,
                self.assignment_type.value,
                self.gene_assignment_type.value,
                self.penalty_score,
                self.isoforms,  # Property will convert from int list if needed
                self.genes)  # Property will convert from int list if needed

    def __setstate__(self, state):
        # When unpickling, store as strings (no pools available during pickle)
        self._string_pools = None
        self.assignment_id = state[0]
        self.read_id = state[1]
        self._chr_id_str = state[2]
        self.chr_id_int = None
        self.start = state[3]
        self.end = state[4]
        self.genomic_region = (state[5], state[6])
        self.multimapper = state[7]
        self.polyA_found = state[8]
        self.assignment_type = ReadAssignmentType(state[9])
        self.gene_assignment_type = ReadAssignmentType(state[10])
        self.penalty_score = state[11]
        self._isoforms = state[12]
        self._genes = state[13]
        self.barcode_id = None
        self.umi_id = None
        self.gene_ids = []
        self.isoform_ids = []

    @classmethod
    def deserialize(cls, infile, string_pools):
        assert string_pools is not None, "string_pools is required"
        read_assignment = cls.__new__(cls)
        read_assignment._string_pools = string_pools
        read_assignment.assignment_id = read_int(infile)
        read_assignment.read_id = read_string(infile)
        read_assignment.chr_id_int = read_int_or_none(infile)
        read_assignment.start = read_int(infile)
        read_assignment.end = read_int(infile)
        read_assignment.genomic_region = (read_int(infile), read_int(infile))
        bool_arr = read_bool_array(infile, 2)
        read_assignment.multimapper = bool_arr[0]
        read_assignment.polyA_found = bool_arr[1]
        read_assignment.assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.gene_assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.penalty_score = float(read_int(infile)) / float(SHORT_FLOAT_MULTIPLIER)
        read_assignment.gene_ids = read_list(infile, read_int)
        read_assignment.isoform_ids = read_list(infile, read_int)
        return read_assignment

    @classmethod
    def deserialize_from_read_assignment(cls, infile, string_pools):
        assert string_pools is not None, "string_pools is required"
        read_assignment = cls.__new__(cls)
        read_assignment._string_pools = string_pools
        read_assignment.assignment_id = read_int(infile)
        read_assignment.read_id = read_string(infile)
        read_assignment.genomic_region = (read_int(infile), read_int(infile))
        exons = read_list_of_pairs(infile, read_int)
        read_assignment.start = exons[0][0]
        read_assignment.end = exons[-1][1]
        read_list_of_pairs(infile, read_int)
        bool_arr = read_bool_array(infile, 3)
        read_assignment.multimapper = bool_arr[0]
        read_assignment.polyA_found = bool_arr[1]
        read_int_neg(infile)
        read_int_neg(infile)
        read_int_neg(infile)
        read_int_neg(infile)
        # Read group: stored as int IDs (dynamic pools loaded from separate file before deserialization)
        read_assignment.read_group_ids = read_list(infile, read_int)
        # Barcode/UMI pools are built in sorted order, so IDs are deterministic
        read_assignment.barcode_id = read_int_or_none(infile)
        read_assignment.umi_id = read_int_or_none(infile)
        read_string(infile)
        read_string(infile)
        read_assignment.chr_id_int = read_int_or_none(infile)
        read_short_int(infile)
        read_assignment.assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.gene_assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.penalty_score = 0.0
        isoform_matches = read_list(infile, lambda f: IsoformMatch.deserialize(f, string_pools))
        gene_set = set()
        isoform_set = set()
        for m in isoform_matches:
            read_assignment.penalty_score = min(read_assignment.penalty_score, isoform_matches[0].penalty_score)
            if m.assigned_gene:
                gene_set.add(m.assigned_gene_id)
            if m.assigned_transcript:
                isoform_set.add(m.assigned_transcript_id)
        read_assignment.gene_ids = list(gene_set)
        read_assignment.isoform_ids = list(isoform_set)

        read_dict(infile)
        read_dict(infile)
        read_short_int(infile)
        read_list(infile, read_int_neg)
        read_list(infile, read_int_neg)
        return read_assignment

    def serialize(self, outfile):
        write_int(self.assignment_id, outfile)
        write_string(self.read_id, outfile)
        write_int_or_none(self.chr_id_int, outfile)
        write_int(self.start, outfile)
        write_int(self.end, outfile)
        write_int(self.genomic_region[0], outfile)
        write_int(self.genomic_region[1], outfile)
        write_bool_array([self.multimapper, self.polyA_found], outfile)
        write_short_int(self.assignment_type.value, outfile)
        write_short_int(self.gene_assignment_type.value, outfile)
        write_int(int(self.penalty_score * SHORT_FLOAT_MULTIPLIER), outfile)
        write_list(self.gene_ids, outfile, write_int)
        write_list(self.isoform_ids, outfile, write_int)


class ReadAssignment:
    """
    Complete read assignment with isoform matches and additional metadata.

    Memory Optimization:
        chr_id, barcode, and umi are stored as integers referencing
        shared string pools (string_pools parameter required).
    """
    assignment_id_generator = SimpleIDDistributor()

    def __init__(self, read_id, assignment_type, string_pools, match=None):
        assert string_pools is not None, "string_pools is required"
        self.assignment_id = ReadAssignment.assignment_id_generator.increment()
        self.read_id = read_id
        self.genomic_region = (0, 0)
        self.exons = None
        self.corrected_exons = None
        self.corrected_introns = None
        self.gene_info = None
        self.multimapper = False
        self.polyA_found = False
        self.cage_found = False
        self.polya_info = None

        # String interning for memory optimization
        self._string_pools = string_pools
        self.read_group_ids = []  # List of integer IDs for read groups
        self.barcode_id = None  # Integer ID for cell/spatial barcode
        self.umi_id = None  # Integer ID for unique molecular identifier
        self.chr_id_int = None  # Integer ID for chromosome

        self.mapped_strand = "."  # Keep as single char (1 byte)
        self.strand = "."  # Keep as single char (1 byte)
        self.mapping_quality = 0
        self.assignment_type = assignment_type
        if match is None:
            self.isoform_matches = []
        elif isinstance(match, list):
            self.isoform_matches = match
        else:
            self.isoform_matches = [match]
        if self.assignment_type == ReadAssignmentType.ambiguous:
            assigned_genes = set([m.assigned_gene for m in self.isoform_matches])
            self.gene_assignment_type = ReadAssignmentType.ambiguous if len(
                assigned_genes) > 1 else ReadAssignmentType.unique
        elif self.assignment_type == ReadAssignmentType.inconsistent_ambiguous:
            assigned_genes = set([m.assigned_gene for m in self.isoform_matches])
            self.gene_assignment_type = ReadAssignmentType.inconsistent_ambiguous if len(
                assigned_genes) > 1 else ReadAssignmentType.inconsistent
        else:
            self.gene_assignment_type = self.assignment_type

        self.additional_info = {}
        self.additional_attributes = {}
        self.introns_match = False
        self.exon_gene_profile = []
        self.intron_gene_profile = []

    @property
    def chr_id(self):
        """Return chromosome string"""
        if self.chr_id_int is not None:
            return self._string_pools.chromosome_pool.get_str(self.chr_id_int)
        return "."

    @chr_id.setter
    def chr_id(self, value):
        """Set chromosome from string"""
        self.chr_id_int = self._string_pools.chromosome_pool.get_int(value) if value else None

    @property
    def barcode(self):
        """Return barcode string"""
        if self.barcode_id is not None:
            return self._string_pools.barcode_pool.get_str(self.barcode_id)
        return None

    @barcode.setter
    def barcode(self, value):
        """Set barcode from string"""
        self.barcode_id = self._string_pools.barcode_pool.get_int(value) if value else None

    @property
    def umi(self):
        """Return UMI string"""
        if self.umi_id is not None:
            return self._string_pools.umi_pool.get_str(self.umi_id)
        return None

    @umi.setter
    def umi(self, value):
        """Set UMI from string"""
        self.umi_id = self._string_pools.umi_pool.get_int(value) if value else None

    @property
    def read_group(self):
        """Return read group as list of strings"""
        if not self.read_group_ids:
            return []
        return self._string_pools.read_group_from_ids(self.read_group_ids)

    @read_group.setter
    def read_group(self, value):
        """Set read group from list of strings"""
        if not value:
            self.read_group_ids = []
        else:
            self.read_group_ids = self._string_pools.read_group_to_ids(value)

    @classmethod
    def deserialize(cls, infile, gene_info, string_pools):
        assert string_pools is not None, "string_pools is required"
        read_assignment = cls.__new__(cls)
        read_assignment._string_pools = string_pools
        read_assignment.assignment_id = read_int(infile)
        read_assignment.read_id = read_string(infile)
        read_assignment.genomic_region = (read_int(infile), read_int(infile))
        read_assignment.exons = read_list_of_pairs(infile, read_int)
        read_assignment.corrected_exons = read_list_of_pairs(infile, read_int)
        read_assignment.corrected_introns = junctions_from_blocks(read_assignment.corrected_exons)
        read_assignment.gene_info = gene_info
        bool_arr = read_bool_array(infile, 3)
        read_assignment.multimapper = bool_arr[0]
        read_assignment.polyA_found = bool_arr[1]
        read_assignment.cage_found = bool_arr[2]
        read_assignment.polya_info = PolyAInfo(read_int_neg(infile), read_int_neg(infile), read_int_neg(infile), read_int_neg(infile))
        # Read group: stored as int IDs (dynamic pools loaded from separate file before deserialization)
        read_assignment.read_group_ids = read_list(infile, read_int)
        # Barcode/UMI pools are built in sorted order, so IDs are deterministic
        read_assignment.barcode_id = read_int_or_none(infile)
        read_assignment.umi_id = read_int_or_none(infile)
        read_assignment.mapped_strand = read_string(infile)
        read_assignment.strand = read_string(infile)
        read_assignment.chr_id_int = read_int_or_none(infile)
        read_assignment.mapping_quality = read_short_int(infile)
        read_assignment.assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.gene_assignment_type = ReadAssignmentType(read_short_int(infile))
        read_assignment.isoform_matches = read_list(infile, lambda f: IsoformMatch.deserialize(f, string_pools))
        read_assignment.additional_info = read_dict(infile)
        read_assignment.additional_attributes = read_dict(infile)
        read_assignment.introns_match = bool(read_short_int(infile))
        read_assignment.exon_gene_profile = read_list(infile, read_int_neg)
        read_assignment.intron_gene_profile = read_list(infile, read_int_neg)
        return read_assignment

    def serialize(self, outfile):
        write_int(self.assignment_id, outfile)
        write_string(self.read_id, outfile)
        write_int(self.genomic_region[0], outfile)
        write_int(self.genomic_region[1], outfile)
        write_list_of_pairs(self.exons, outfile, write_int)
        write_list_of_pairs(self.corrected_exons, outfile, write_int)
        write_bool_array([self.multimapper, self.polyA_found, self.cage_found], outfile)
        write_int_neg(self.polya_info.external_polya_pos, outfile)
        write_int_neg(self.polya_info.external_polyt_pos, outfile)
        write_int_neg(self.polya_info.internal_polya_pos, outfile)
        write_int_neg(self.polya_info.internal_polyt_pos, outfile)
        # Serialize read_group as int IDs (dynamic pools saved to separate file)
        write_list(self.read_group_ids, outfile, write_int)
        # Barcode/UMI pools are built in sorted order, so IDs are deterministic
        write_int_or_none(self.barcode_id, outfile)
        write_int_or_none(self.umi_id, outfile)
        write_string(self.mapped_strand, outfile)
        write_string(self.strand, outfile)
        write_int_or_none(self.chr_id_int, outfile)
        write_short_int(self.mapping_quality, outfile)
        write_short_int(self.assignment_type.value, outfile)
        write_short_int(self.gene_assignment_type.value, outfile)
        write_list(self.isoform_matches, outfile, IsoformMatch.serialize)
        write_dict(self.additional_info, outfile)
        write_dict(self.additional_attributes, outfile)
        write_short_int(int(self.introns_match), outfile)
        write_list(self.exon_gene_profile, outfile, write_int_neg)
        write_list(self.intron_gene_profile, outfile, write_int_neg)

    def add_match(self, match):
        self.isoform_matches.append(match)

    def set_assignment_type(self, assignment_type):
        self.assignment_type = assignment_type

    def start(self):
        return self.exons[0][0]

    def end(self):
        return self.exons[-1][1]

    def length(self):
        return sum([x[1] - x[0] + 1 for x in self.exons])

    def exon_count(self):
        return len(self.exons)

    def set_additional_info(self, key, value):
        self.additional_info[key] = value

    def set_additional_attribute(self, key, value):
        self.additional_attributes[key] = value

    def add_match_attribute(self, match_event):
        for m in self.isoform_matches:
            m.add_subclassification(match_event)


match_subtype_printable_names = \
    {MatchEventSubtype.ism_left : ('ism_5', 'ism_3', 'ism'),
     MatchEventSubtype.ism_right : ('ism_3', 'ism_5', 'ism'),
     MatchEventSubtype.exon_elongation_left : ('exon_elongation_5', 'exon_elongation_3', 'exon_elongation'),
     MatchEventSubtype.exon_elongation_right : ('exon_elongation_3', 'exon_elongation_5', 'exon_elongation'),
     MatchEventSubtype.major_exon_elongation_left: ('major_exon_elongation_5', 'major_exon_elongation_3', 'major_exon_elongation'),
     MatchEventSubtype.major_exon_elongation_right: ('major_exon_elongation_3', 'major_exon_elongation_5', 'major_exon_elongation'),
     MatchEventSubtype.fake_terminal_exon_left : ('fake_terminal_exon_5', 'fake_terminal_exon_3', 'fake_terminal_exon'),
     MatchEventSubtype.fake_terminal_exon_right : ('fake_terminal_exon_3', 'fake_terminal_exon_5', 'fake_terminal_exon'),
     MatchEventSubtype.terminal_exon_misalignment_left : ('terminal_exon_misalignment_5', 'terminal_exon_misalignment_3', 'terminal_exon_misalignment'),
     MatchEventSubtype.terminal_exon_misalignment_right : ('terminal_exon_misalignment_3', 'terminal_exon_misalignment_5', 'terminal_exon_misalignment'),
     MatchEventSubtype.incomplete_intron_retention_left: ('incomplete_intron_retention_5', 'incomplete_intron_retention_3', 'incomplete_intron_retention'),
     MatchEventSubtype.incomplete_intron_retention_right: ('incomplete_intron_retention_3', 'incomplete_intron_retention_5', 'incomplete_intron_retention'),
     MatchEventSubtype.extra_intron_flanking_left: ('extra_intron_5', 'extra_intron_3', 'extra_intron'),
     MatchEventSubtype.extra_intron_flanking_right: ('extra_intron_3', 'extra_intron_5', 'extra_intron'),
     MatchEventSubtype.alt_left_site_known: ('alt_donor_site_known', 'alt_acceptor_site_known', 'alternative_splice_site_known'),
     MatchEventSubtype.alt_right_site_known: ('alt_acceptor_site_known', 'alt_donor_site_known', 'alternative_splice_site_known'),
     MatchEventSubtype.alt_left_site_novel: ('alt_donor_site_novel', 'alt_acceptor_site_novel', 'alternative_splice_site_novel'),
     MatchEventSubtype.alt_right_site_novel: ('alt_acceptor_site_novel', 'alt_donor_site_novel', 'alternative_splice_site_novel'),
     MatchEventSubtype.terminal_site_match_left: ('tss_match', 'tes_match', 'terminal_position_match'),
     MatchEventSubtype.terminal_site_match_right: ('tes_match', 'tss_match', 'terminal_position_match'),
     MatchEventSubtype.terminal_site_match_left_precise: ('tss_match_precise', 'tes_match_precise', 'terminal_position_match_precise'),
     MatchEventSubtype.terminal_site_match_right_precise: ('tes_match_precise', 'tss_match_precise', 'terminal_position_match_precise')
     }
     #MatchEventSubtype.alternative_polya_site_left: ('alternative_polya_site_5', 'alternative_polya_site_3'),
     #MatchEventSubtype.alternative_polya_site_right: ('alternative_polya_site_3', 'alternative_polya_site_5'),
     #MatchEventSubtype.internal_polya_left: ('internal_polya_site_5', 'internal_polya_site_3'),
     #MatchEventSubtype.internal_polya_right: ('internal_polya_site_3', 'internal_polya_site_5'),
     #MatchEventSubtype.correct_polya_site_left: ('correct_polya_site_5', 'correct_polya_site_3'),
     #MatchEventSubtype.correct_polya_site_right: ('correct_polya_site_3', 'correct_polya_site_5'),
     #MatchEventSubtype.alternative_tss_left: ('alternative_tss_5', 'alternative_tss_3'),
     #MatchEventSubtype.alternative_tss_right: ('alternative_tss_3', 'alternative_tss_5')}


def match_subtype_to_str(event, strand):
    event_subtype = event.event_type
    if event_subtype in match_subtype_printable_names.keys():
        if strand is None:
            logger.warning("Strand is not set for site-dependent modifications")
        if strand == '-':
            return match_subtype_printable_names[event_subtype][1]
        elif strand == '+':
            return match_subtype_printable_names[event_subtype][0]
        else:
            return match_subtype_printable_names[event_subtype][2]
    return event_subtype.name


def regions_to_str(regions):
    return ",".join([str(x[0]) + "-" + str(x[1]) for x in regions])


def match_subtype_to_str_with_additional_info(event, strand, read_introns, isoform_introns):
    event_subtype = event.event_type
    additional_info = ""
    if event_subtype in {MatchEventSubtype.intron_retention,
                         MatchEventSubtype.unspliced_intron_retention,
                         MatchEventSubtype.incomplete_intron_retention_left,
                         MatchEventSubtype.incomplete_intron_retention_right,
                         MatchEventSubtype.fake_micro_intron_retention}:
        if event.isoform_region != SupplementaryMatchConstants.undefined_region:
            introns = isoform_introns[event.isoform_region[0]:event.isoform_region[1]+1]
            additional_info = ":" + regions_to_str(introns)
    elif event_subtype in {MatchEventSubtype.major_exon_elongation_left, MatchEventSubtype.major_exon_elongation_right,
                           MatchEventSubtype.exon_elongation_left, MatchEventSubtype.exon_elongation_right,
                           MatchEventSubtype.alternative_tss_left, MatchEventSubtype.alternative_tss_right,
                           MatchEventSubtype.alternative_polya_site_left, MatchEventSubtype.alternative_polya_site_right,
                           MatchEventSubtype.correct_polya_site_right, MatchEventSubtype.correct_polya_site_left,
                           MatchEventSubtype.internal_polya_left, MatchEventSubtype.internal_polya_right,
                           MatchEventSubtype.terminal_site_match_left, MatchEventSubtype.terminal_site_match_right,
                           MatchEventSubtype.terminal_site_match_left_precise,
                           MatchEventSubtype.terminal_site_match_right_precise}:
        # elongation events
        additional_info = ":" + str(event.event_info)
    else:
        if event.read_region != SupplementaryMatchConstants.undefined_region and \
                event.read_region[0] >= 0 and event.read_region[1] >= 0:
            introns = read_introns[event.read_region[0]:event.read_region[1]+1]
            additional_info = ":" + regions_to_str(introns)

    return match_subtype_to_str(event, strand) + additional_info


def is_matching_assignment(isoform_assignment):
    if isoform_assignment.assignment_type == ReadAssignmentType.unique:
        return True
    elif isoform_assignment.assignment_type.is_unique():
        allowed_set = {MatchEventSubtype.none,
                       MatchEventSubtype.fsm,
                       MatchEventSubtype.exon_misalignment,
                       MatchEventSubtype.intron_shift,
                       MatchEventSubtype.terminal_site_match_left,
                       MatchEventSubtype.terminal_site_match_right,
                       MatchEventSubtype.terminal_site_match_left_precise,
                       MatchEventSubtype.terminal_site_match_right_precise,
                       MatchEventSubtype.correct_polya_site_right,
                       MatchEventSubtype.correct_polya_site_left,
                       MatchEventSubtype.exon_elongation_left,
                       MatchEventSubtype.exon_elongation_right}

        return all(m.event_type in allowed_set for m in isoform_assignment.isoform_matches[0].match_subclassifications)
    return False



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Basic k-mer indexers for barcode calling.

KmerIndexer: Dictionary-based, best for small to medium sets
ArrayKmerIndexer: Array-based with 2-bit encoding, best for k <= 8
"""

import math
from typing import List, Tuple, Iterable, Dict, DefaultDict
from collections import defaultdict


class KmerIndexer:
    """
    Basic k-mer indexer using dictionary-based storage.

    Indexes sequences by their k-mers for fast approximate matching.
    Best for small to medium barcode sets.
    """

    def __init__(self, known_strings: Iterable[str], kmer_size: int = 6):
        """
        Initialize k-mer index.

        Args:
            known_strings: Collection of reference sequences (barcodes/UMIs)
            kmer_size: Length of k-mers to use for indexing
        """
        self.seq_list: List[str] = list(known_strings)
        self.k: int = kmer_size
        self.index: DefaultDict[str, List[int]] = defaultdict(list)
        self._index()

    def _get_kmers(self, seq: str) -> Iterable[str]:
        """Generate all k-mers from a sequence using sliding window."""
        if len(seq) < self.k:
            return
        kmer = seq[:self.k]
        yield kmer
        for i in range(self.k, len(seq)):
            kmer = kmer[1:] + seq[i]
            yield kmer

    def _index(self) -> None:
        """Build k-mer index from all sequences."""
        for i, barcode in enumerate(self.seq_list):
            for kmer in self._get_kmers(barcode):
                self.index[kmer].append(i)

    def append(self, barcode: str) -> None:
        """Add a new barcode to the index."""
        self.seq_list.append(barcode)
        index = len(self.seq_list) - 1
        for kmer in self._get_kmers(barcode):
            self.index[kmer].append(index)

    def empty(self) -> bool:
        """Check if index is empty."""
        return len(self.seq_list) == 0

    def get_occurrences(self, sequence: str, max_hits: int = 0, min_kmers: int = 1,
                       hits_delta: int = 1, ignore_equal: bool = False) -> List[Tuple[str, int, List[int]]]:
        """
        Find indexed sequences with shared k-mers.

        Args:
            sequence: Query sequence to search
            max_hits: Maximum number of results (0 = unlimited)
            min_kmers: Minimum shared k-mers required
            hits_delta: Include results within this many k-mers of top hit
            ignore_equal: Skip exact matches

        Returns:
            List of (sequence, shared_kmer_count, kmer_positions) tuples,
            sorted by shared k-mer count (descending)
        """
        barcode_counts: DefaultDict[int, int] = defaultdict(int)
        barcode_positions: DefaultDict[int, List[int]] = defaultdict(list)

        for pos, kmer in enumerate(self._get_kmers(sequence)):
            for i in self.index[kmer]:
                barcode_counts[i] += 1
                barcode_positions[i].append(pos)

        result = []
        for i in barcode_counts.keys():
            count = barcode_counts[i]
            if count < min_kmers:
                continue
            if ignore_equal and self.seq_list[i] == sequence:
                continue
            result.append((self.seq_list[i], count, barcode_positions[i]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = list(sorted(result, reverse=True, key=lambda x: x[1]))

        if max_hits == 0:
            return result
        return result[:max_hits]

    def _get_kmers_substr(self, seq, start, end):
        if end - start + 1 < self.k:
            return
        kmer = seq[start:start + self.k]
        yield kmer
        for i in range(start + self.k, min(len(seq), end + 1)):
            kmer = kmer[1:] + seq[i]
            yield kmer

    # [start, end]
    def get_occurrences_substr(self, sequence, start, end, max_hits=0, min_kmers=1, hits_delta=1) -> List[Tuple[str, int, List[int]]]:
        """
        Find indexed sequences with shared k-mers.

        Args:
            sequence: Query sequence to search
            max_hits: Maximum number of results (0 = unlimited)
            min_kmers: Minimum shared k-mers required
            hits_delta: Include results within this many k-mers of top hit
            ignore_equal: Skip exact matches

        Returns:
            List of (sequence, shared_kmer_count, kmer_positions) tuples,
            sorted by shared k-mer count (descending)
        """
        barcode_counts: DefaultDict[int, int] = defaultdict(int)
        barcode_positions: DefaultDict[int, List[int]] = defaultdict(list)

        for pos, kmer in enumerate(self._get_kmers_substr(sequence, start, end)):
            for i in self.index[kmer]:
                barcode_counts[i] += 1
                barcode_positions[i].append(pos)

        result = []
        for i in barcode_counts.keys():
            count = barcode_counts[i]
            if count < min_kmers:
                continue
            result.append((self.seq_list[i], count, barcode_positions[i]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = list(sorted(result, reverse=True, key=lambda x: x[1]))

        if max_hits == 0:
            return result
        return result[:max_hits]


class ArrayKmerIndexer:
    """
    Optimized k-mer indexer using array-based storage and binary encoding.

    Converts k-mers to integers (2 bits per nucleotide) for faster lookup.
    Memory usage: O(4^k) array entries. Best for k <= 8.
    """

    # Nucleotide to 2-bit encoding (A=00, C=01, G=10, T=11)
    NUCL2BIN: Dict[str, int] = {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'a': 0, 'c': 1, 'g': 2, 't': 3}

    def __init__(self, known_strings: Iterable[str], kmer_size: int = 6):
        """
        Initialize array-based k-mer index.

        Args:
            known_strings: Collection of reference sequences
            kmer_size: Length of k-mers (recommended <= 8)
        """
        self.seq_list: List[str] = list(known_strings)
        self.k: int = kmer_size
        total_kmers = int(math.pow(4, kmer_size))  # 4^k possible k-mers
        self.index: List[List[int]] = [[] for _ in range(total_kmers)]
        self.mask: int = (1 << (2 * self.k)) - 1
        self._index()

    def _get_kmer_indexes(self, seq: str) -> Iterable[int]:
        """Generate binary-encoded k-mer indices using sliding window."""
        if len(seq) < self.k:
            return

        # Initialize first k-mer
        kmer_idx = 0
        for i in range(self.k):
            kmer_idx |= ArrayKmerIndexer.NUCL2BIN[seq[i]] << ((self.k - i - 1) * 2)
        yield kmer_idx

        # Slide window: shift left 2 bits, add new nucleotide
        for i in range(self.k, len(seq)):
            kmer_idx <<= 2
            kmer_idx &= self.mask
            kmer_idx |= ArrayKmerIndexer.NUCL2BIN[seq[i]]
            yield kmer_idx

    def _index(self) -> None:
        """Build k-mer index from all sequences."""
        for i, barcode in enumerate(self.seq_list):
            for kmer_idx in self._get_kmer_indexes(barcode):
                self.index[kmer_idx].append(i)

    def append(self, barcode: str) -> None:
        """Add a new barcode to the index."""
        self.seq_list.append(barcode)
        index = len(self.seq_list) - 1
        for kmer_idx in self._get_kmer_indexes(barcode):
            self.index[kmer_idx].append(index)

    def empty(self) -> bool:
        """Check if index is empty."""
        return len(self.seq_list) == 0

    def get_occurrences(self, sequence: str, max_hits: int = 0, min_kmers: int = 1,
                       hits_delta: int = 1, ignore_equal: bool = False) -> List[Tuple[str, int, List[int]]]:
        """Find indexed sequences with shared k-mers."""
        barcode_counts: DefaultDict[int, int] = defaultdict(int)
        barcode_positions: DefaultDict[int, List[int]] = defaultdict(list)

        for pos, kmer_idx in enumerate(self._get_kmer_indexes(sequence)):
            for i in self.index[kmer_idx]:
                barcode_counts[i] += 1
                barcode_positions[i].append(pos)

        result = []
        for i in barcode_counts.keys():
            count = barcode_counts[i]
            if count < min_kmers:
                continue
            if ignore_equal and self.seq_list[i] == sequence:
                continue
            result.append((self.seq_list[i], count, barcode_positions[i]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = list(sorted(result, reverse=True, key=lambda x: x[1]))

        if max_hits == 0:
            return result
        return result[:max_hits]



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2021-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import queue
from collections import defaultdict

from .common import find_closest, overlaps

logger = logging.getLogger('IsoQuant')

VERTEX_polya = -10
VERTEX_read_end = -11
VERTEX_polyt = -20
VERTEX_read_start = -21


def is_terminal_vertex(v):
    return v[0] in [VERTEX_polya, VERTEX_read_end]


def is_starting_vertex(v):
    return v[0] in [VERTEX_polyt, VERTEX_read_start]


class IntronCollector:
    def __init__(self, gene_info, delta=0):
        self.gene_info = gene_info
        self.known_introns = set(self.gene_info.intron_profiles.features)
        self.delta = delta
        # all intron counts
        # clustered intron counts
        self.clustered_introns = defaultdict(int)
        # how introns were corrected after clustering
        self.intron_correction_map = {}
        self.discarded_introns = set()

    def collect_introns(self, read_assignments):
        all_introns = defaultdict(int)
        for assignment in read_assignments:
            if not assignment.corrected_introns or assignment.multimapper:
                continue
            for intron in assignment.corrected_introns:
                all_introns[intron] += 1
        return all_introns

    def construct_similar_intron_map(self, all_introns):
        ordered_introns = sorted(all_introns.keys())
        similar_intron_map = defaultdict(list)
        # create a dict of similar introns
        for i, intron in enumerate(ordered_introns):
            j = i + 1
            while j < len(ordered_introns) and abs(ordered_introns[j][0] - intron[0]) <= self.delta:
                if abs(ordered_introns[j][1] - intron[1]) <= self.delta:
                    similar_intron_map[intron].append(ordered_introns[j])
                    similar_intron_map[ordered_introns[j]].append(intron)
                j += 1
        return similar_intron_map

    def cluster_introns(self, all_introns, min_count):
        similar_intron_map = self.construct_similar_intron_map(all_introns)
        introns_sorted_by_counts = sorted([(v, k) for k, v in all_introns.items()], reverse=True)
        for count, intron in introns_sorted_by_counts:
            if intron in self.known_introns:
                # known intron is always added as trustworthy
                self.clustered_introns[intron] = count
            elif intron in similar_intron_map:
                # intron has a similar intron
                similar_introns = []
                for similar_intron in similar_intron_map[intron]:
                    if similar_intron in self.clustered_introns:
                        # if similar intron was already added to the cluster with the higher count
                        similar_introns.append((count, similar_intron))

                if similar_introns:
                    # take the best one as a substitute
                    substitute_intron = sorted(similar_introns, reverse=True)[0]
                    self.clustered_introns[substitute_intron[1]] += substitute_intron[0]
                    self.intron_correction_map[intron] = substitute_intron[1]
                else:
                    # no introns were found
                    self.clustered_introns[intron] = count
            elif count < min_count:
                self.discarded_introns.add(intron)
                continue
            else:
                self.clustered_introns[intron] = count

    def process(self, read_assignments, min_count):
        logger.debug("Processing introns")
        all_introns = self.collect_introns(read_assignments)
        logger.debug(all_introns)
        self.cluster_introns(all_introns, min_count)
        logger.debug(self.clustered_introns)

    def add_substitute(self, original_intron, substitute_intron):
        self.clustered_introns[substitute_intron] += self.clustered_introns[original_intron]
        del self.clustered_introns[original_intron]
        self.intron_correction_map[original_intron] = substitute_intron

    def discard(self, intron):
        self.discarded_introns.add(intron)
        if intron in self.clustered_introns:
            del self.clustered_introns[intron]

    def substitute(self, v):
        if v in self.intron_correction_map:
            v = self.intron_correction_map[v]
        return v

    def simplify_correction_map(self):
        all_introns = sorted(self.intron_correction_map.keys())
        to_remove = set()
        for intron in all_introns:
            subs = self.intron_correction_map[intron]
            if subs in self.discarded_introns:
                to_remove.add(intron)
                continue
            if subs not in self.intron_correction_map:
                continue
            while subs in self.intron_correction_map:
                subs = self.intron_correction_map[subs]
            if subs in self.discarded_introns:
                to_remove.add(intron)
                continue
            self.intron_correction_map[intron] = subs

        for intron in to_remove:
            self.discard(intron)
            del self.intron_correction_map[intron]


class IntronGraph:
    def __init__(self, params, gene_info, read_assignments):
        self.params = params
        self.gene_info = gene_info
        self.read_assignments = read_assignments

        self.incoming_edges = defaultdict(set)
        self.outgoing_edges = defaultdict(set)
        self.intron_collector = IntronCollector(gene_info, params.delta)
        self.max_coverage = 0
        self.edge_weights = defaultdict(int)

        self.starting_known_positions = defaultdict(list)
        self.terminal_known_positions = defaultdict(list)
        for t, introns in self.gene_info.all_isoforms_introns.items():
            if not introns:
                continue
            self.starting_known_positions[introns[0]].append(self.gene_info.all_isoforms_exons[t][0][0])
            self.terminal_known_positions[introns[-1]].append(self.gene_info.all_isoforms_exons[t][-1][1])

        # logger.debug("Collecting introns for %s" % self.gene_info.gene_db_list[0].id)
        self.intron_collector.process(read_assignments, self.params.min_novel_intron_count)
        self.construct()
        if self.params.debug:
            self.print_graph()
        self.simplify()
        self.attach_terminal_positions()
        if self.params.debug:
            self.print_graph()

    def add_edge(self, v1, v2):
        if v1 in self.intron_collector.intron_correction_map:
            v1 = self.intron_collector.intron_correction_map[v1]
        if v2 in self.intron_collector.intron_correction_map:
            v2 = self.intron_collector.intron_correction_map[v2]
        self.outgoing_edges[v1].add(v2)
        self.incoming_edges[v2].add(v1)
        self.edge_weights[(v1, v2)] += 1

    def is_isolated(self, v):
        return not self.outgoing_edges[v] and not self.incoming_edges[v]

    def is_monointron(self, v):
        no_outgoing = not self.outgoing_edges[v] or \
                      (all(is_terminal_vertex(x) for x in self.outgoing_edges[v]))
        no_incoming = not self.incoming_edges[v] or (
            all(is_starting_vertex(x) for x in self.incoming_edges[v]))
        return no_incoming and no_outgoing

    def signleton_dead_end(self, v):
        path = set()
        while self.intron_collector.clustered_introns[v] == 1:
            path.add(v)
            if len(self.outgoing_edges[v]) != 1:
                break
            v = list(self.outgoing_edges[v])[0]

        if len(self.outgoing_edges[v]) == 0:
            return path
        return set()

    def signleton_dead_start(self, v):
        path = set()
        while self.intron_collector.clustered_introns[v] == 1:
            path.add(v)
            if len(self.incoming_edges[v]) != 1:
                break
            v = list(self.incoming_edges[v])[0]

        if len(self.incoming_edges[v]) == 0:
            return path
        return set()

    def get_outgoing(self, intron, v_type=None):
        res = []
        if intron not in self.outgoing_edges:
            return res
        if v_type is None:
            for v in self.outgoing_edges[intron]:
                if v[0] >= 0:
                    res.append(v)
        else:
            for v in self.outgoing_edges[intron]:
                if v[0] == v_type:
                    res.append(v)
        return sorted(res)

    def get_incoming(self, intron, v_type=None):
        res = []
        if intron not in self.incoming_edges:
            return res
        if v_type is None:
            for v in self.incoming_edges[intron]:
                if v[0] >= 0:
                    res.append(v)
        else:
            for v in self.incoming_edges[intron]:
                if v[0] == v_type:
                    res.append(v)
        return sorted(res)

    # merge vertex to its substitute, remove if isolated
    def collapse_vertex(self, to_collapse, substitute_vertex):
        self.outgoing_edges[substitute_vertex].update(self.outgoing_edges[to_collapse])
        for i in self.outgoing_edges[to_collapse]:
            self.incoming_edges[i].remove(to_collapse)
            self.incoming_edges[i].add(substitute_vertex)

        self.incoming_edges[substitute_vertex].update(self.incoming_edges[to_collapse])
        for i in self.incoming_edges[to_collapse]:
            self.outgoing_edges[i].remove(to_collapse)
            self.outgoing_edges[i].add(substitute_vertex)

        self.intron_collector.add_substitute(to_collapse, substitute_vertex)

    def construct(self):
        logger.debug("Constructing for %d" % self.gene_info.start)
        for assignment in self.read_assignments:
            if assignment.multimapper or any(intron in self.intron_collector.discarded_introns for intron in assignment.corrected_introns):
                continue

            for i in range(len(assignment.corrected_introns) - 1):
                intron1 = assignment.corrected_introns[i]
                intron2 = assignment.corrected_introns[i + 1]
                self.add_edge(intron1, intron2)

        self.max_coverage =  max(self.intron_collector.clustered_introns.values()) if self.intron_collector.clustered_introns else 0

    def simplify(self):
        logger.debug("Simplifying graph")
        self.clean_tips_and_bulges()
        self.remove_singleton_dead_ends()
        self.remove_isolates()
        self.intron_collector.simplify_correction_map()

    def clean_tips_and_bulges(self):
        # check all outgoing edges
        to_remove = set()
        logger.debug("Removing outgoing tips and bulges")
        for current_intron in sorted(self.outgoing_edges.keys()):
            out_introns = self.outgoing_edges[current_intron]
            substitute_dict = self.collapse_vertex_set(out_introns)
            for i in sorted(substitute_dict.keys()):
                if i in to_remove:
                    continue
                to_remove.add(i)
                self.collapse_vertex(i, substitute_dict[i])

        for i in to_remove:
            del self.outgoing_edges[i]
            del self.incoming_edges[i]
        to_remove.clear()

        # check all incoming edges
        logger.debug("Removing incoming tips and bulges")
        for current_intron in sorted(self.incoming_edges.keys()):
            inc_introns = self.incoming_edges[current_intron]
            substitute_dict = self.collapse_vertex_set(inc_introns)
            for i in sorted(substitute_dict.keys()):
                if i in to_remove:
                    continue
                to_remove.add(i)
                self.collapse_vertex(i, substitute_dict[i])

        for i in to_remove:
            del self.outgoing_edges[i]
            del self.incoming_edges[i]

    def remove_singleton_dead_ends(self):
        to_clean = {}
        for current_intron in sorted(self.outgoing_edges.keys()):
            if self.intron_collector.clustered_introns[current_intron] < self.params.singleton_adjacent_cov:
                # singleton removal can only be used for high-covered introns
                continue
            # collect singleton paths (coverage 1, only one outgoing edge)
            outgoing_paths = [self.signleton_dead_end(i) for i in self.outgoing_edges[current_intron]]
            if any(len(p) == 0 for p in outgoing_paths):
                continue
            to_clean[current_intron] = set()
            for p in outgoing_paths:
                to_clean[current_intron].update(p)

        for intron in to_clean.keys():
            self.outgoing_edges[intron] = set()
            for i in to_clean[intron]:
                if i in self.outgoing_edges:
                    del self.outgoing_edges[i]
                if i in self.incoming_edges:
                    del self.incoming_edges[i]

        to_clean = {}
        for current_intron in sorted(self.incoming_edges.keys()):
            if self.intron_collector.clustered_introns[current_intron] < self.params.singleton_adjacent_cov:
                # singleton removal can only be used for high-covered introns
                continue
            incoming_paths = [self.signleton_dead_start(i) for i in self.incoming_edges[current_intron]]
            if any(len(p) == 0 for p in incoming_paths):
                continue
            to_clean[current_intron] = set()
            for p in incoming_paths:
                to_clean[current_intron].update(p)

        for intron in to_clean.keys():
            self.incoming_edges[intron] = set()
            for i in to_clean[intron]:
                if i in self.outgoing_edges:
                    del self.outgoing_edges[i]
                if i in self.incoming_edges:
                    del self.incoming_edges[i]

    def remove_isolates(self):
        to_remove = set()
        # check all isolated vertices
        isolated = set()
        logger.debug("Collapsing isolated introns")
        for intron in self.intron_collector.clustered_introns.keys():
            if self.is_isolated(intron):
                isolated.add(intron)
        substitute_dict = self.collapse_vertex_set(isolated)
        for i in sorted(substitute_dict.keys()):
            to_remove.add(i)
            self.collapse_vertex(i, substitute_dict[i])

        for i in to_remove:
            del self.outgoing_edges[i]
            del self.incoming_edges[i]
        to_remove.clear()
        # self.print_graph()

        # remove low covered isolated vertices
        logger.debug("Removing isolated introns")
        count_cutoff = self.params.min_novel_isolated_intron_abs
        for intron in isolated:
            if intron not in self.intron_collector.clustered_introns or intron in self.intron_collector.known_introns:
                # already removed or known
                continue
            if self.intron_collector.clustered_introns[intron] < count_cutoff:
                logger.debug("Removing isolated %s : %d" % (str(intron), self.intron_collector.clustered_introns[intron]))
                self.intron_collector.discard(intron)
                to_remove.add(intron)

        for i in to_remove:
            del self.outgoing_edges[i]
            del self.incoming_edges[i]

    def collapse_vertex_set(self, vertex_set):
        if len(vertex_set) <= 1:
            return {}

        approved_set = set()
        substitute_dict = {}
        # get counts
        vertex_counts = sorted([(self.intron_collector.clustered_introns[i], i) for i in vertex_set], reverse=True)
        for count, vertex in vertex_counts:
            similar_vertices = []
            for i in sorted(approved_set):
                start_dist = abs(i[0] - vertex[0])
                end_dist = abs(i[1] - vertex[1])
                if start_dist < self.params.graph_clustering_distance and \
                        end_dist < self.params.graph_clustering_distance and \
                        count < self.intron_collector.clustered_introns[i] * self.params.graph_clustering_ratio:
                    # found similar intron among approved
                    similar_vertices.append((start_dist + end_dist, i))

            if not similar_vertices:
                # no similar introns
                approved_set.add(vertex)
            else:
                # selecting the most similar intron among approved
                substitute_vertex = sorted(similar_vertices)[0][1]
                substitute_dict[vertex] = substitute_vertex

        return substitute_dict

    def attach_terminal_positions(self):
        # logger.debug("Setting terminal positions paths for %s" % self.gene_info.gene_db_list[0].id)
        polya_ends, read_ends, polyt_starts, read_starts = self.collect_terminal_positions()

        for intron in sorted(self.intron_collector.clustered_introns):
            self.attach_transcpt_ends(intron, polya_ends, read_ends, read_end=True)
            self.attach_transcpt_ends(intron, polyt_starts, read_starts, read_end=False)

    def attach_transcpt_ends(self, intron, polya_confirmed_positions, read_terminal_positions, read_end=True):
        read_ends_cutoff = self.params.terminal_position_abs
        logger.debug(str(intron) + " => " + str(polya_confirmed_positions[intron]))
        clustered_polyas = self.cluster_polya_positions(polya_confirmed_positions[intron], intron, read_end)
        if clustered_polyas:
            read_ends_cutoff = max(read_ends_cutoff, max(clustered_polyas.values()) * self.params.terminal_position_rel)
            extra_end_positions = {}
            furtherst_confirmed_position = max(clustered_polyas.keys()) if read_end else min(clustered_polyas.keys())
            for position, count in read_terminal_positions[intron].items():
                if read_end and position >= furtherst_confirmed_position + self.params.apa_delta:
                    extra_end_positions[position] = count
                elif not read_end and position <= furtherst_confirmed_position - self.params.apa_delta :
                    extra_end_positions[position] = count
        else:
            extra_end_positions = read_terminal_positions[intron]

        if read_end and intron in self.outgoing_edges and len(self.outgoing_edges[intron]) > 0:
            # intron has outgoing edges, hard cut off
            neighboring_cov = max(self.intron_collector.clustered_introns[i] for i in self.outgoing_edges[intron])
            read_ends_cutoff = max(read_ends_cutoff, neighboring_cov * self.params.terminal_internal_position_rel)
        elif not read_end and intron in self.incoming_edges and len(self.incoming_edges[intron]) > 0:
            # intron has incoming edges, hard cut off
            neighboring_cov = max(self.intron_collector.clustered_introns[i] for i in self.incoming_edges[intron])
            read_ends_cutoff = max(read_ends_cutoff, neighboring_cov * self.params.terminal_internal_position_rel)

        logger.debug(str(intron) + " +> " + str(extra_end_positions))

        terminal_positions = self.cluster_terminal_positions(extra_end_positions,
                                                             read_end=read_end,
                                                             cutoff=read_ends_cutoff)
        logger.debug("POLYAs clustered:")
        logger.debug(clustered_polyas)
        logger.debug("Teminal clustered:")
        logger.debug(terminal_positions)
        if read_end:
            # if intron in self.terminal_known_positions:
            #    logger.debug("Annotated terminal positions: " + str(sorted(self.terminal_known_positions[intron])))
            # logger.debug("PolyA terminal positions: " + str(sorted(clustered_polyas.keys())))
            # logger.debug("Simple terminal positions: " + str(sorted(terminal_positions.keys())))
            for pos in clustered_polyas.keys():
                self.outgoing_edges[intron].add((VERTEX_polya, pos))
            for pos in terminal_positions.keys():
                self.outgoing_edges[intron].add((VERTEX_read_end, pos))
        else:
            # if intron in self.starting_known_positions:
            #    logger.debug("Annotated terminal positions: " + str(sorted(self.starting_known_positions[intron])))
            # logger.debug("PolyA terminal positions: " + str(sorted(clustered_polyas.keys())))
            # logger.debug("Simple terminal positions: " + str(sorted(terminal_positions.keys())))
            for pos in clustered_polyas.keys():
                self.incoming_edges[intron].add((VERTEX_polyt, pos))
            for pos in terminal_positions.keys():
                self.incoming_edges[intron].add((VERTEX_read_start, pos))

    def collect_terminal_positions(self):
        polya_ends = defaultdict(lambda: defaultdict(int))
        read_ends = defaultdict(lambda: defaultdict(int))
        polyt_starts = defaultdict(lambda: defaultdict(int))
        read_starts = defaultdict(lambda: defaultdict(int))

        for assignment in self.read_assignments:
            if assignment.multimapper or not assignment.corrected_introns:
                continue
            if any(intron in self.intron_collector.discarded_introns for intron in
                   assignment.corrected_introns):
                continue

            starting_intron = self.intron_collector.substitute(assignment.corrected_introns[0])
            read_start = assignment.corrected_exons[0][0]
            if read_start >= starting_intron[0]:
                # corner case when substituted intron appears to be shifted to the left
                continue
            polyt_detected = assignment.strand == '-' and \
                             (assignment.polya_info.external_polyt_pos != -1 or
                              assignment.polya_info.internal_polyt_pos != -1)
            if polyt_detected:
                polyt_starts[starting_intron][read_start] += 1
            elif not self.is_start_internal(starting_intron, read_start):
                read_starts[starting_intron][read_start] += 1

            terminating_intron = self.intron_collector.substitute(assignment.corrected_introns[-1])
            read_end = assignment.corrected_exons[-1][1]
            if read_end <= terminating_intron[1]:
                # corner case when substituted intron appears to be shifted to the right
                continue
            polya_detected = assignment.strand == '+' and \
                             (assignment.polya_info.external_polya_pos != -1 or
                              assignment.polya_info.internal_polya_pos != -1)
            if polya_detected:
                polya_ends[terminating_intron][read_end] += 1
            elif not self.is_end_internal(terminating_intron, read_end):
                read_ends[terminating_intron][read_end] += 1

        # self.print_terminal(read_starts, "Read starts")
        # self.print_terminal(polyt_starts, "Read polyT")
        # self.print_terminal(read_ends, "Read ends")
        # self.print_terminal(polya_ends, "Read polyA")

        return polya_ends, read_ends, polyt_starts, read_starts

    def cluster_polya_positions(self, position_dict, intron, read_end):
        clustered_counts = {}
        if not position_dict:
            return clustered_counts

        known_positions = self.terminal_known_positions[intron] if read_end else self.starting_known_positions[intron]
        while position_dict:
            best_pair = max(position_dict.items(), key=lambda x:x[1])
            top_position = best_pair[0]
            nearest_position, diff_to_nearest_position = find_closest(top_position, known_positions)
            if nearest_position and diff_to_nearest_position <= self.params.apa_delta:
                logger.debug("Using reference TSS/TES: %d -> %d" % (top_position, nearest_position))
                top_position = nearest_position
            if read_end:
                assert top_position > intron[1]
            else:
                assert top_position < intron[0]
            total_count = 0
            for pos in range(top_position - self.params.apa_delta, top_position + self.params.apa_delta + 1):
                if pos in position_dict:
                    total_count += position_dict[pos]
                    del position_dict[pos]
            clustered_counts[top_position] = total_count

        max_count = max(clustered_counts.values())
        if max_count == self.params.terminal_position_abs:
            return clustered_counts
        cutoff = max(max_count * self.params.terminal_position_rel, self.params.terminal_position_abs)
        result = {}
        for k,v in clustered_counts.items():
            if v >= cutoff:
                result[k] = v
        return result

    def cluster_terminal_positions(self, position_dict, read_end, cutoff=0):
        # simple solution for now
        if not position_dict:
            return {}

        total_count = sum(position_dict.values())
        pos = max(position_dict.keys()) if read_end else min(position_dict.keys())
        if total_count < cutoff:
            return {}
        return {pos: total_count}

    def is_start_internal(self, intron, read_start):
        is_internal = False
        # checking previous introns
        for inc in self.incoming_edges[intron]:
            if inc[1] - self.params.delta <= read_start:
                is_internal = True
                break
        return is_internal

    def is_end_internal(self, intron, read_end):
        is_internal = False
        # checking previous introns
        for out in self.outgoing_edges[intron]:
            if out[0] + self.params.delta >= read_end:
                is_internal = True
                break
        return is_internal

    def get_connected_component(self, path):
        intron_queue = queue.Queue()
        processed_set = set()
        for intron in path:
            if intron[0] < 0:
                continue
            intron_queue.put_nowait(intron)
            processed_set.add(intron)

        while not intron_queue.empty():
            intron = intron_queue.get_nowait()
            for i in self.outgoing_edges[intron]:
                if i[0] >= 0 and i not in processed_set:
                    processed_set.add(i)
                    intron_queue.put_nowait(i)
            for i in self.incoming_edges[intron]:
                if i[0] >= 0 and i not in processed_set:
                    processed_set.add(i)
                    intron_queue.put_nowait(i)
        return processed_set

    def get_max_component_coverage(self, path):
        component = self.get_connected_component(path)
        if not component:
            return 0
        return max(self.intron_collector.clustered_introns[i] for i in component)

    def get_overlapping_component_max_coverage(self, coordinates):
        processed_introns = set()
        all_vertices = set(self.outgoing_edges.keys())
        all_vertices.update(self.incoming_edges.keys())
        for intron in all_vertices:
            if intron in processed_introns:
                continue

            max_right_pos = intron[1]
            for v in self.outgoing_edges[intron]:
                if is_terminal_vertex(v):
                    max_right_pos = max(max_right_pos, v[1])
                else:
                    max_right_pos = max(max_right_pos, v[0])
            min_left_pos = intron[0]
            for v in self.incoming_edges[intron]:
                min_left_pos = min(min_left_pos, v[1])

            if overlaps(coordinates, (min_left_pos, max_right_pos)):
                processed_introns.update(self.get_connected_component([intron]))

        if not processed_introns:
            return 0
        return max(self.intron_collector.clustered_introns[i] for i in processed_introns)

    def print_graph(self):
        logger.debug("Printing graph")
        logger.debug("Vertices: %d, substituted: %d, discarded: %d" % (len(self.intron_collector.clustered_introns),
                                                                       len(self.intron_collector.intron_correction_map),
                                                                       len(self.intron_collector.discarded_introns)))
        logger.debug("Outgoing: %d, incoming %d" % (len(self.incoming_edges), len(self.outgoing_edges)))
        logger.debug("Discarded")
        logger.debug(self.intron_collector.discarded_introns)
        logger.debug("Collected")
        logger.debug(self.intron_collector.clustered_introns)
        logger.debug("Substituted")
        logger.debug(self.intron_collector.intron_correction_map)
        for intron in sorted(self.intron_collector.clustered_introns.keys()):
            logger.debug("Intron %s, count %d -> %s" % (str(intron), self.intron_collector.clustered_introns[intron],
                                                        ",".join([str(x) + ":" + str(self.edge_weights[(intron, x)])
                                                                  for x in self.outgoing_edges[intron]])))

        for intron in sorted(self.intron_collector.clustered_introns.keys()):
            logger.debug("Intron %s, count %d <- %s" % (str(intron), self.intron_collector.clustered_introns[intron],
                                                        ",".join([str(x) for x in self.incoming_edges[intron]])))

    def print_terminal(self, pos_dict, s=""):
        logger.debug(s)
        for intron in pos_dict:
            logger.debug("Intron %s -> %s" % (str(intron),
                                              ",".join([str(x) + ":" + str(pos_dict[intron][x])
                                                        for x in sorted(pos_dict[intron].keys())])))



# ############################################################################
# # Copyright (c) 2022-2026 University of Helsinki
# # Copyright (c) 2019-2022 Saint Petersburg State University
# # # All Rights Reserved
# # See file LICENSE for details.
# ############################################################################
#
# import unittest
# from unittest.mock import Mock
# from src.intron_graph import *
# from src.gene_info import GeneInfo
# from src.isoform_assignment import ReadAssignment, ReadAssignmentType, IsoformMatch, PolyAInfo
# from src.common import junctions_from_blocks
#
#
# class MockParams:
#     """Mock parameters object for testing"""
#     def __init__(self):
#         self.delta = 6
#         self.graph_clustering_distance = 10
#         self.min_novel_intron_count = 2
#         self.min_novel_isolated_intron_abs = 3
#         self.min_novel_isolated_intron_rel = 0.02
#         self.terminal_position_abs = 1
#         self.terminal_position_rel = 0.05
#         self.singleton_adjacent_cov = 100
#
#
# def create_mock_gene_info(exons_list, chr_id="chr1"):
#     """
#     Create a mock GeneInfo object with given exons.
#     exons_list: list of tuples representing exon coordinates
#     """
#     gene_info = Mock(spec=GeneInfo)
#     gene_info.chr_id = chr_id
#     gene_info.start = min(e[0] for e in exons_list)
#     gene_info.end = max(e[1] for e in exons_list)
#     gene_info.all_isoforms_introns = {}
#     gene_info.all_isoforms_exons = {"transcript1": exons_list}
#
#     # Create intron profiles
#     introns = junctions_from_blocks(exons_list)
#     gene_info.intron_profiles = Mock()
#     gene_info.intron_profiles.features = introns
#
#     return gene_info
#
#
# def create_read_assignment(read_id, exon_blocks, chr_id="chr1", strand="+",
#                           assignment_type=ReadAssignmentType.unique, polyA_found=False):
#     """
#     Create a ReadAssignment object for testing.
#     """
#     match = IsoformMatch(None, None, None)
#     match.assigned_gene = "gene1"
#     match.assigned_transcript = "transcript1"
#
#     read_assignment = ReadAssignment(read_id, assignment_type, match)
#     read_assignment.chr_id = chr_id
#     read_assignment.strand = strand
#     read_assignment.mapped_strand = strand
#     read_assignment.exons = exon_blocks
#     read_assignment.corrected_exons = exon_blocks
#     read_assignment.corrected_introns = junctions_from_blocks(exon_blocks)
#     read_assignment.genomic_region = (exon_blocks[0][0], exon_blocks[-1][1])
#     read_assignment.polyA_found = polyA_found
#     read_assignment.polya_info = PolyAInfo(-1, -1, -1, -1)
#     read_assignment.read_group = "test_group"
#     read_assignment.mapping_quality = 60
#
#     return read_assignment
#
#
# class TestIntronCollector(unittest.TestCase):
#     """Test the IntronCollector class"""
#
#     def setUp(self):
#         self.exons = [(100, 200), (300, 400), (500, 600)]
#         self.gene_info = create_mock_gene_info(self.exons)
#         self.params = MockParams()
#
#     def test_collector(self):
#         """Test IntronCollector initialization"""
#         collector = IntronCollector(self.gene_info, self.params)
#         self.assertEqual(collector.gene_info, self.gene_info)
#         self.assertEqual(collector.delta, self.params.delta)
#         self.assertIsNotNone(collector.known_introns)
#
#     def test_collect_introns_simple(self):
#         """Test collecting introns from reads matching known structure"""
#         collector = IntronCollector(self.gene_info, self.params)
#
#         # Create reads matching the gene structure
#         read1 = create_read_assignment("read1", self.exons)
#         read2 = create_read_assignment("read2", self.exons)
#
#         read_assignments = [read1, read2]
#         all_introns = collector.collect_introns(read_assignments)
#
#         # Should have collected the known introns
#         expected_introns = junctions_from_blocks(self.exons)
#         for intron in expected_introns:
#             self.assertIn(intron, all_introns)
#
#     def test_collect_novel_introns(self):
#         """Test collecting novel introns"""
#         collector = IntronCollector(self.gene_info, self.params)
#
#         # Create read with novel intron
#         novel_exons = [(100, 200), (350, 400), (500, 600)]  # Different middle exon
#         read1 = create_read_assignment("read1", novel_exons)
#         read2 = create_read_assignment("read2", novel_exons)
#         read3 = create_read_assignment("read3", novel_exons)
#
#         read_assignments = [read1, read2, read3]
#         all_introns = collector.collect_introns(read_assignments)
#
#         # Should have collected the novel intron
#         novel_introns = junctions_from_blocks(novel_exons)
#         for intron in novel_introns:
#             self.assertIn(intron, all_introns)
#
#
# class TestIntronGraph(unittest.TestCase):
#     """Test the IntronGraph class"""
#
#     def setUp(self):
#         self.exons = [(100, 200), (300, 400), (500, 600)]
#         self.gene_info = create_mock_gene_info(self.exons)
#         self.params = MockParams()
#
#     def test_initialization(self):
#         """Test IntronGraph initialization"""
#         read_assignments = []
#         graph = IntronGraph(read_assignments, self.gene_info, self.params)
#
#         self.assertEqual(graph.gene_info, self.gene_info)
#         self.assertEqual(graph.params, self.params)
#         self.assertIsNotNone(graph.intron_collector)
#
#     def test_simple_linear_graph(self):
#         """Test graph construction for simple linear transcript"""
#         # Create reads matching the gene structure
#         reads = [
#             create_read_assignment("read1", self.exons),
#             create_read_assignment("read2", self.exons),
#             create_read_assignment("read3", self.exons),
#         ]
#
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#
#         # Graph should have vertices for each intron boundary
#         introns = junctions_from_blocks(self.exons)
#         self.assertGreater(len(graph.outgoing_edges), 0)
#
#     def test_graph_with_alternative_splicing(self):
#         """Test graph construction with alternative splicing"""
#         # Create reads with different exon structures
#         variant1_exons = [(100, 200), (300, 400), (500, 600)]
#         variant2_exons = [(100, 200), (350, 450), (500, 600)]  # Alternative middle exon
#
#         reads = [
#             create_read_assignment("read1", variant1_exons),
#             create_read_assignment("read2", variant1_exons),
#             create_read_assignment("read3", variant2_exons),
#             create_read_assignment("read4", variant2_exons),
#         ]
#
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#
#         # Graph should have branches for alternative splicing
#         self.assertGreater(len(graph.outgoing_edges), 0)
#
#     def test_monoexonic_transcript(self):
#         """Test graph with monoexonic (single exon) transcript"""
#         mono_exons = [(100, 500)]
#         mono_gene_info = create_mock_gene_info(mono_exons)
#
#         reads = [
#             create_read_assignment("read1", mono_exons),
#             create_read_assignment("read2", mono_exons),
#         ]
#
#         graph = IntronGraph(reads, mono_gene_info, self.params)
#         graph.construct()
#
#         # Monoexonic should have minimal graph structure
#         # No introns, so graph should be simple or empty
#         self.assertIsNotNone(graph.outgoing_edges)
#
#     def test_add_edge(self):
#         """Test adding edges to the graph"""
#         reads = [create_read_assignment("read1", self.exons)]
#         graph = IntronGraph(reads, self.gene_info, self.params)
#
#         # Add an edge
#         v1 = (200, 'R')  # Right end of first exon
#         v2 = (300, 'L')  # Left end of second exon
#         graph.add_edge(v1, v2)
#
#         self.assertIn(v1, graph.outgoing_edges)
#         self.assertIn(v2, graph.outgoing_edges[v1])
#         self.assertEqual(graph.edge_weights[(v1, v2)], 1)
#
#     def test_get_outgoing_edges(self):
#         """Test retrieving outgoing edges"""
#         reads = [
#             create_read_assignment("read1", self.exons),
#             create_read_assignment("read2", self.exons),
#         ]
#
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#
#         # Find a vertex with outgoing edges
#         if graph.outgoing_edges:
#             vertex = list(graph.outgoing_edges.keys())[0]
#             outgoing = graph.get_outgoing(vertex)
#             self.assertIsInstance(outgoing, list)
#
#     def test_is_isolated(self):
#         """Test checking if a vertex is isolated"""
#         reads = [create_read_assignment("read1", self.exons)]
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#
#         # Create an isolated vertex
#         isolated_vertex = (9999, 'L')
#         result = graph.is_isolated(isolated_vertex)
#         # An isolated vertex has no connections
#         self.assertTrue(result or isolated_vertex not in graph.outgoing_edges)
#
#
# class TestTerminalVertexFunctions(unittest.TestCase):
#     """Test terminal vertex helper functions"""
#
#     def test_is_terminal_vertex(self):
#         """Test identifying terminal vertices"""
#         self.assertTrue(is_terminal_vertex(VERTEX_polya))
#         self.assertTrue(is_terminal_vertex(VERTEX_read_end))
#         self.assertFalse(is_terminal_vertex((100, 'L')))
#
#     def test_is_starting_vertex(self):
#         """Test identifying starting vertices"""
#         self.assertTrue(is_starting_vertex(VERTEX_polyt))
#         self.assertTrue(is_starting_vertex(VERTEX_read_start))
#         self.assertFalse(is_starting_vertex((100, 'R')))
#
#
# class TestComplexGraphScenarios(unittest.TestCase):
#     """Test complex graph construction scenarios"""
#
#     def setUp(self):
#         self.params = MockParams()
#
#     def test_multiple_isoforms(self):
#         """Test graph with multiple isoforms from same gene"""
#         # Create gene with two isoforms
#         isoform1_exons = [(100, 200), (300, 400), (500, 600), (700, 800)]
#         isoform2_exons = [(100, 200), (300, 400), (700, 800)]  # Skips middle exon
#
#         gene_info = create_mock_gene_info(isoform1_exons)
#
#         reads = [
#             create_read_assignment("read1", isoform1_exons),
#             create_read_assignment("read2", isoform1_exons),
#             create_read_assignment("read3", isoform2_exons),
#             create_read_assignment("read4", isoform2_exons),
#         ]
#
#         graph = IntronGraph(reads, gene_info, self.params)
#         graph.construct()
#
#         # Should have vertices for both isoforms
#         self.assertGreater(len(graph.outgoing_edges), 0)
#
#     def test_reads_with_polya(self):
#         """Test graph construction with polyA-containing reads"""
#         exons = [(100, 200), (300, 400), (500, 600)]
#         gene_info = create_mock_gene_info(exons)
#
#         reads = [
#             create_read_assignment("read1", exons, polyA_found=True),
#             create_read_assignment("read2", exons, polyA_found=True),
#             create_read_assignment("read3", exons, polyA_found=False),
#         ]
#
#         graph = IntronGraph(reads, gene_info, self.params)
#         graph.construct()
#
#         # PolyA reads should influence terminal positions
#         self.assertIsNotNone(graph.terminal_known_positions)
#
#     def test_partial_read_coverage(self):
#         """Test reads that don't cover full transcript"""
#         full_exons = [(100, 200), (300, 400), (500, 600), (700, 800)]
#         gene_info = create_mock_gene_info(full_exons)
#
#         # Reads covering different parts
#         reads = [
#             create_read_assignment("read1", [(100, 200), (300, 400)]),  # First half
#             create_read_assignment("read2", [(500, 600), (700, 800)]),  # Second half
#             create_read_assignment("read3", full_exons),  # Full length
#         ]
#
#         graph = IntronGraph(reads, gene_info, self.params)
#         graph.construct()
#
#         # Should handle partial coverage
#         self.assertIsNotNone(graph.outgoing_edges)
#
#     def test_noisy_introns(self):
#         """Test graph with low-coverage (noisy) introns"""
#         main_exons = [(100, 200), (300, 400), (500, 600)]
#         gene_info = create_mock_gene_info(main_exons)
#
#         # Most reads follow main pattern
#         reads = [create_read_assignment(f"read{i}", main_exons) for i in range(10)]
#
#         # One read has a noisy intron
#         noisy_exons = [(100, 200), (320, 380), (500, 600)]
#         reads.append(create_read_assignment("noisy_read", noisy_exons))
#
#         graph = IntronGraph(reads, gene_info, self.params)
#         graph.construct()
#
#         # Noisy intron should be filtered or have low weight
#         self.assertIsNotNone(graph.edge_weights)
#
#
# class TestGraphSimplification(unittest.TestCase):
#     """Test graph simplification methods"""
#
#     def setUp(self):
#         self.params = MockParams()
#         self.exons = [(100, 200), (300, 400), (500, 600)]
#         self.gene_info = create_mock_gene_info(self.exons)
#
#     def test_simplify_graph(self):
#         """Test graph simplification"""
#         reads = [create_read_assignment(f"read{i}", self.exons) for i in range(5)]
#
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#         graph.simplify()
#
#         # After simplification, graph should still be valid
#         self.assertIsNotNone(graph.outgoing_edges)
#
#     def test_remove_low_coverage_edges(self):
#         """Test removal of low-coverage edges"""
#         main_exons = [(100, 200), (300, 400), (500, 600)]
#
#         # High coverage for main path
#         reads = [create_read_assignment(f"read{i}", main_exons) for i in range(20)]
#
#         # Low coverage alternative
#         alt_exons = [(100, 200), (350, 450), (500, 600)]
#         reads.append(create_read_assignment("alt_read", alt_exons))
#
#         graph = IntronGraph(reads, self.gene_info, self.params)
#         graph.construct()
#
#         # Low coverage edges may be removed during simplification
#         initial_edge_count = len(graph.edge_weights)
#         graph.simplify()
#         # Simplification may reduce edges
#         self.assertLessEqual(len(graph.edge_weights), initial_edge_count + 10)
#
#
# if __name__ == '__main__':
#     unittest.main()



############################################################################
# Copyright (c) 2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

"""Unit tests for CI test config loading (YAML and TSV formats)."""

import os
import sys
import tempfile
import shutil
import unittest

import yaml

# Add tests/github to path so we can import the modules
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "github"))

from run_pipeline import (
    load_tsv_config as pipeline_load_tsv,
    load_yaml_config as pipeline_load_yaml,
    load_config as pipeline_load_config,
    check_value,
)
from run_barcode_test import (
    load_config as barcode_load_config,
)
from cfg2yaml import (
    detect_barcode_mode,
    convert_cfg_to_yaml,
    load_baseline_file,
)
from update_defaults import (
    update_yaml_baselines,
    load_tsv_config as defaults_load_tsv,
)


class TestLoadTsvConfig(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_basic_loading(self):
        path = os.path.join(self.test_dir, "test.cfg")
        with open(path, "w") as f:
            f.write("name\tmy_test\n")
            f.write("output\t/tmp/out\n")
            f.write("datatype\tont\n")
        result = pipeline_load_tsv(path)
        self.assertEqual(result["name"], "my_test")
        self.assertEqual(result["output"], "/tmp/out")
        self.assertEqual(result["datatype"], "ont")

    def test_comments_skipped(self):
        path = os.path.join(self.test_dir, "test.cfg")
        with open(path, "w") as f:
            f.write("# this is a comment\n")
            f.write("name\tmy_test\n")
            f.write("# another comment\n")
        result = pipeline_load_tsv(path)
        self.assertEqual(len(result), 1)
        self.assertEqual(result["name"], "my_test")

    def test_blank_lines_skipped(self):
        path = os.path.join(self.test_dir, "test.cfg")
        with open(path, "w") as f:
            f.write("name\tmy_test\n")
            f.write("\n")
            f.write("output\t/tmp\n")
        result = pipeline_load_tsv(path)
        self.assertEqual(len(result), 2)

    def test_missing_file_returns_empty(self):
        result = pipeline_load_tsv("/nonexistent/path.cfg")
        self.assertEqual(result, {})


class TestLoadYamlConfig(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def _write_yaml(self, data: dict, name: str = "test.yaml") -> str:
        path = os.path.join(self.test_dir, name)
        with open(path, "w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)
        return path

    def test_basic_loading(self):
        path = self._write_yaml({
            "name": "my_test",
            "output": "/tmp/out",
            "datatype": "ont",
        })
        config, baselines = pipeline_load_yaml(path)
        self.assertEqual(config["name"], "my_test")
        self.assertEqual(config["output"], "/tmp/out")
        self.assertEqual(baselines, {})

    def test_baselines_extracted(self):
        path = self._write_yaml({
            "name": "my_test",
            "baselines": {
                "transcripts": {"full_recall": 84.7, "full_precision": 96.6},
                "performance": {"max_rss": 1234.0},
            },
        })
        config, baselines = pipeline_load_yaml(path)
        self.assertNotIn("baselines", config)
        self.assertIn("transcripts", baselines)
        self.assertAlmostEqual(baselines["transcripts"]["full_recall"], 84.7)
        self.assertIn("performance", baselines)

    def test_values_converted_to_strings(self):
        path = self._write_yaml({
            "name": "test",
            "threads": 8,
            "tolerance": 0.01,
        })
        config, _ = pipeline_load_yaml(path)
        self.assertEqual(config["threads"], "8")
        self.assertEqual(config["tolerance"], "0.01")

    def test_no_baselines_key(self):
        path = self._write_yaml({"name": "test"})
        config, baselines = pipeline_load_yaml(path)
        self.assertEqual(baselines, {})

    def test_missing_file_returns_empty(self):
        config, baselines = pipeline_load_yaml("/nonexistent/path.yaml")
        self.assertEqual(config, {})
        self.assertEqual(baselines, {})

    def test_empty_file(self):
        path = os.path.join(self.test_dir, "empty.yaml")
        with open(path, "w") as f:
            f.write("")
        config, baselines = pipeline_load_yaml(path)
        self.assertEqual(config, {})
        self.assertEqual(baselines, {})


class TestLoadConfigDispatch(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_yaml_extension(self):
        path = os.path.join(self.test_dir, "test.yaml")
        with open(path, "w") as f:
            yaml.dump({"name": "yaml_test", "baselines": {"transcripts": {"r": 1.0}}}, f)
        config, baselines = pipeline_load_config(path)
        self.assertEqual(config["name"], "yaml_test")
        self.assertIn("transcripts", baselines)

    def test_yml_extension(self):
        path = os.path.join(self.test_dir, "test.yml")
        with open(path, "w") as f:
            yaml.dump({"name": "yml_test"}, f)
        config, baselines = pipeline_load_config(path)
        self.assertEqual(config["name"], "yml_test")
        self.assertEqual(baselines, {})

    def test_cfg_extension(self):
        path = os.path.join(self.test_dir, "test.cfg")
        with open(path, "w") as f:
            f.write("name\tcfg_test\n")
        config, baselines = pipeline_load_config(path)
        self.assertEqual(config["name"], "cfg_test")
        self.assertEqual(baselines, {})

    def test_barcode_load_config_yaml(self):
        path = os.path.join(self.test_dir, "bc.yaml")
        with open(path, "w") as f:
            yaml.dump({
                "name": "bc_test",
                "mode": "custom_sc",
                "baselines": {"barcode": {"precision": 99.5}},
            }, f)
        config, baselines = barcode_load_config(path)
        self.assertEqual(config["name"], "bc_test")
        self.assertAlmostEqual(baselines["barcode"]["precision"], 99.5)


class TestCheckValue(unittest.TestCase):
    def test_within_tolerance(self):
        self.assertEqual(check_value(100.0, 100.5, "metric"), 0)

    def test_below_tolerance(self):
        self.assertNotEqual(check_value(100.0, 90.0, "metric"), 0)

    def test_above_tolerance(self):
        self.assertNotEqual(check_value(100.0, 110.0, "metric"), 0)

    def test_exact_match(self):
        self.assertEqual(check_value(50.0, 50.0, "metric"), 0)

    def test_negative_exact_match(self):
        self.assertEqual(check_value(-1.0, -1.0, "metric"), 0)

    def test_negative_mismatch(self):
        self.assertNotEqual(check_value(-1.0, -2.0, "metric"), 0)

    def test_custom_tolerance(self):
        # 20% tolerance: 100 * 0.8 = 80, so 85 should pass
        self.assertEqual(check_value(100.0, 85.0, "metric", 0.2), 0)
        # but 70 should fail
        self.assertNotEqual(check_value(100.0, 70.0, "metric", 0.2), 0)


class TestDetectBarcodeMode(unittest.TestCase):
    def test_pipeline_config(self):
        config = {"genome": "/path/fa", "genedb": "/path/gtf", "datatype": "ont"}
        self.assertFalse(detect_barcode_mode(config))

    def test_barcode_config(self):
        config = {"mode": "custom_sc", "molecule": "10x.mdf", "qa_mode": "tenX_v3"}
        self.assertTrue(detect_barcode_mode(config))

    def test_sc_pipeline_config(self):
        # SC pipeline has both barcode-like isoquant_options and pipeline keys
        # but barcode keys are inside isoquant_options, not top-level
        config = {"genome": "/path/fa", "genedb": "/path/gtf", "datatype": "ont",
                  "isoquant_options": '"-m tenX_v3"'}
        self.assertFalse(detect_barcode_mode(config))

    def test_mixed_keys_pipeline_wins(self):
        # If both barcode and pipeline keys are present, it's a pipeline config
        config = {"genome": "/path/fa", "genedb": "/path/gtf", "mode": "custom_sc"}
        self.assertFalse(detect_barcode_mode(config))


class TestCfg2Yaml(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_pipeline_conversion(self):
        # Create cfg
        cfg_path = os.path.join(self.test_dir, "test.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\tmy_pipeline\n")
            f.write("genome\t/path/to/genome.fa\n")
            f.write("genedb\t/path/to/genes.gtf\n")
            f.write("datatype\tont\n")
            f.write("etalon\ttest.etl\n")

        # Create etalon file
        etl_path = os.path.join(self.test_dir, "test.etl")
        with open(etl_path, "w") as f:
            f.write("full_recall\t84.70\n")
            f.write("full_precision\t96.60\n")

        yaml_path, yaml_data = convert_cfg_to_yaml(cfg_path)
        self.assertEqual(yaml_path, os.path.join(self.test_dir, "test.yaml"))
        self.assertEqual(yaml_data["name"], "my_pipeline")
        self.assertNotIn("etalon", yaml_data)
        self.assertIn("baselines", yaml_data)
        self.assertAlmostEqual(yaml_data["baselines"]["transcripts"]["full_recall"], 84.7)
        self.assertAlmostEqual(yaml_data["baselines"]["transcripts"]["full_precision"], 96.6)

    def test_barcode_conversion(self):
        cfg_path = os.path.join(self.test_dir, "bc.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\tmy_barcode\n")
            f.write("mode\tcustom_sc\n")
            f.write("molecule\t10x.mdf\n")
            f.write("etalon\tbc.etl\n")

        etl_path = os.path.join(self.test_dir, "bc.etl")
        with open(etl_path, "w") as f:
            f.write("precision\t99.89\n")
            f.write("recall\t86.82\n")

        yaml_path, yaml_data = convert_cfg_to_yaml(cfg_path)
        self.assertNotIn("etalon", yaml_data)
        self.assertIn("baselines", yaml_data)
        self.assertIn("barcode", yaml_data["baselines"])
        self.assertAlmostEqual(yaml_data["baselines"]["barcode"]["precision"], 99.89)

    def test_force_barcode_mode(self):
        cfg_path = os.path.join(self.test_dir, "test.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\ttest\n")
            f.write("etalon\ttest.etl\n")

        etl_path = os.path.join(self.test_dir, "test.etl")
        with open(etl_path, "w") as f:
            f.write("metric1\t1.0\n")

        _, yaml_data = convert_cfg_to_yaml(cfg_path, force_barcode=True)
        self.assertIn("barcode", yaml_data["baselines"])
        self.assertNotIn("transcripts", yaml_data.get("baselines", {}))

    def test_missing_etalon_no_baselines(self):
        cfg_path = os.path.join(self.test_dir, "test.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\ttest\n")
            f.write("genome\t/path/genome.fa\n")
            f.write("datatype\tont\n")

        _, yaml_data = convert_cfg_to_yaml(cfg_path)
        self.assertNotIn("baselines", yaml_data)


class TestLoadBaselineFile(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_basic(self):
        path = os.path.join(self.test_dir, "test.etl")
        with open(path, "w") as f:
            f.write("recall\t84.70\n")
            f.write("precision\t96.60\n")
        result = load_baseline_file(path)
        self.assertAlmostEqual(result["recall"], 84.7)
        self.assertAlmostEqual(result["precision"], 96.6)

    def test_missing_file(self):
        result = load_baseline_file("/nonexistent/file.etl")
        self.assertEqual(result, {})

    def test_comments_and_blanks(self):
        path = os.path.join(self.test_dir, "test.etl")
        with open(path, "w") as f:
            f.write("# header\n")
            f.write("metric1\t1.0\n")
            f.write("\n")
            f.write("metric2\t2.0\n")
        result = load_baseline_file(path)
        self.assertEqual(len(result), 2)


class TestUpdateYamlBaselines(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.run_name = "my_test"
        # Create pipeline output structure
        self.pipeline_dir = os.path.join(self.test_dir, "output", self.run_name)
        os.makedirs(os.path.join(self.pipeline_dir, "gffcompare"))

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def _write_yaml_config(self, data: dict) -> str:
        path = os.path.join(self.test_dir, "config.yaml")
        with open(path, "w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)
        return path

    def _write_tsv(self, path: str, data: dict) -> None:
        with open(path, "w") as f:
            for k, v in data.items():
                f.write("%s\t%.2f\n" % (k, v))

    def test_pipeline_baselines_updated(self):
        config_path = self._write_yaml_config({
            "name": self.run_name,
            "baselines": {
                "transcripts": {"full_recall": 80.0, "full_precision": 90.0},
            },
        })
        # Write new values
        self._write_tsv(
            os.path.join(self.pipeline_dir, "gffcompare", "new_gtf_etalon.tsv"),
            {"full_recall": 85.0, "full_precision": 95.0},
        )

        base_output_dir = os.path.join(self.test_dir, "output")
        update_yaml_baselines(config_path, self.run_name, base_output_dir)

        with open(config_path) as f:
            updated = yaml.safe_load(f)
        self.assertAlmostEqual(updated["baselines"]["transcripts"]["full_recall"], 85.0)
        self.assertAlmostEqual(updated["baselines"]["transcripts"]["full_precision"], 95.0)

    def test_barcode_baselines_updated(self):
        config_path = self._write_yaml_config({
            "name": self.run_name,
            "baselines": {
                "barcode": {"precision": 99.0, "recall": 85.0},
            },
        })
        # Barcode etalon lives at {base_output_dir}/{run_name}.new_etalon.tsv
        base_output_dir = os.path.join(self.test_dir, "output")
        self._write_tsv(
            os.path.join(base_output_dir, self.run_name + ".new_etalon.tsv"),
            {"precision": 99.5, "recall": 87.0},
        )

        update_yaml_baselines(config_path, self.run_name, base_output_dir)

        with open(config_path) as f:
            updated = yaml.safe_load(f)
        self.assertAlmostEqual(updated["baselines"]["barcode"]["precision"], 99.5)
        self.assertAlmostEqual(updated["baselines"]["barcode"]["recall"], 87.0)

    def test_missing_output_skipped(self):
        config_path = self._write_yaml_config({
            "name": self.run_name,
            "baselines": {
                "transcripts": {"full_recall": 80.0},
                "performance": {"max_rss": 1000.0},
            },
        })
        # Only write transcripts output, not performance
        self._write_tsv(
            os.path.join(self.pipeline_dir, "gffcompare", "new_gtf_etalon.tsv"),
            {"full_recall": 85.0},
        )

        base_output_dir = os.path.join(self.test_dir, "output")
        update_yaml_baselines(config_path, self.run_name, base_output_dir)

        with open(config_path) as f:
            updated = yaml.safe_load(f)
        # transcripts updated
        self.assertAlmostEqual(updated["baselines"]["transcripts"]["full_recall"], 85.0)
        # performance unchanged
        self.assertAlmostEqual(updated["baselines"]["performance"]["max_rss"], 1000.0)

    def test_non_baseline_config_preserved(self):
        config_path = self._write_yaml_config({
            "name": self.run_name,
            "output": "/some/path",
            "datatype": "ont",
            "baselines": {
                "transcripts": {"full_recall": 80.0},
            },
        })
        self._write_tsv(
            os.path.join(self.pipeline_dir, "gffcompare", "new_gtf_etalon.tsv"),
            {"full_recall": 85.0},
        )

        base_output_dir = os.path.join(self.test_dir, "output")
        update_yaml_baselines(config_path, self.run_name, base_output_dir)

        with open(config_path) as f:
            updated = yaml.safe_load(f)
        self.assertEqual(updated["name"], self.run_name)
        self.assertEqual(updated["output"], "/some/path")
        self.assertEqual(updated["datatype"], "ont")


class TestYamlTsvRoundTrip(unittest.TestCase):
    """Verify that YAML and TSV configs produce equivalent results when loaded."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_config_values_match(self):
        # Create a TSV config
        cfg_path = os.path.join(self.test_dir, "test.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\tmy_test\n")
            f.write("output\t/tmp/output\n")
            f.write("run_type\ttranscripts,quantification_known\n")
            f.write("genome\t/path/to/genome.fa\n")
            f.write("datatype\tont\n")
            f.write('isoquant_options\t"-t 8 --complete_genedb --force"\n')

        # Create equivalent YAML config
        yaml_path = os.path.join(self.test_dir, "test.yaml")
        with open(yaml_path, "w") as f:
            yaml.dump({
                "name": "my_test",
                "output": "/tmp/output",
                "run_type": "transcripts,quantification_known",
                "genome": "/path/to/genome.fa",
                "datatype": "ont",
                "isoquant_options": '"-t 8 --complete_genedb --force"',
            }, f, default_flow_style=False, sort_keys=False)

        cfg_config, cfg_baselines = pipeline_load_config(cfg_path)
        yaml_config, yaml_baselines = pipeline_load_config(yaml_path)

        for key in cfg_config:
            self.assertEqual(cfg_config[key], yaml_config.get(key),
                             "Mismatch for key %s" % key)

    def test_isoquant_options_processing(self):
        """Verify isoquant_options string processes identically after loading."""
        cfg_path = os.path.join(self.test_dir, "test.cfg")
        with open(cfg_path, "w") as f:
            f.write("name\ttest\n")
            f.write('isoquant_options\t"-t 8 --complete_genedb --force"\n')

        yaml_path = os.path.join(self.test_dir, "test.yaml")
        with open(yaml_path, "w") as f:
            yaml.dump({
                "name": "test",
                "isoquant_options": '"-t 8 --complete_genedb --force"',
            }, f)

        cfg_config, _ = pipeline_load_config(cfg_path)
        yaml_config, _ = pipeline_load_config(yaml_path)

        cfg_opts = cfg_config["isoquant_options"].replace('"', '').split()
        yaml_opts = yaml_config["isoquant_options"].replace('"', '').split()
        self.assertEqual(cfg_opts, yaml_opts)


if __name__ == "__main__":
    unittest.main()



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
10x Genomics barcode detectors.

Supports 10x Genomics v3 single-cell and Visium HD spatial platforms.
"""

import logging
from typing import List, Optional

from .. import ArrayKmerIndexer, Array2BitKmerIndexer
from ..indexers import KmerIndexer
from ..common import (
    find_polyt_start, reverese_complement,
    find_candidate_with_max_score_ssw, find_candidate_with_max_score_ssw_var_len,
    detect_exact_positions, str_to_2bit,
    find_optimal_kmer_size,
)
from .base import TenXBarcodeDetectionResult

logger = logging.getLogger('IsoQuant')


class TenXBarcodeDetector:
    """10x Genomics v3 single-cell barcode detector."""

    TSO = "CCCATGTACTCTGCGTTGATACCACTGCTT"
    R1 = "CTACACGACGCTCTTCCGATCT"  # 10x 3'
    BARCODE_LEN_10X = 16
    UMI_LEN = 12

    UMI_LEN_DELTA = 2
    TERMINAL_MATCH_DELTA = 2
    STRICT_TERMINAL_MATCH_DELTA = 1

    def __init__(self, barcode_list: List[str]):
        """
        Initialize 10x detector.

        Args:
            barcode_list: List of known barcodes (whitelist)
            umi_list: Optional list of known UMIs for validation
        """
        self.r1_indexer = KmerIndexer([TenXBarcodeDetector.R1], kmer_size=7)
        bc_length = len(barcode_list[0])
        bc_count = len(barcode_list)
        if bc_count > 1000000:
            logger.warning("The number of barcodes is large: barcode calling may take substantial amount of time and RAM, and affect the accuracy")
            logger.warning("We suggest to use a sub-list of barcodes derived from short-read analysis whenever possible")

        self.k = find_optimal_kmer_size(bc_length, bc_count)
        if len(barcode_list) <= 100000:
            self.barcode_indexer = ArrayKmerIndexer(barcode_list, kmer_size=self.k)
            self.max_barcodes_hits = 20
            self.min_matching_kmers = 1
            self.min_score = 14
            self.score_diff = 0
        else:
            barcode_bit_list = [str_to_2bit(b) for b in barcode_list]
            self.barcode_indexer = Array2BitKmerIndexer(barcode_bit_list, kmer_size=self.k, seq_len=self.BARCODE_LEN_10X)
            self.max_barcodes_hits = 10
            self.min_matching_kmers = 2
            self.min_score = 16
            self.score_diff = 1

        logger.info("Indexed %d barcodes of length %d with k-mer size %d" % (len(barcode_list), bc_length, self.k))
        logger.info("Minimal alignment score set to %d" % self.min_score)

    def find_barcode_umi(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        """
        Detect barcode and UMI in a read sequence.

        Args:
            read_id: Read identifier
            sequence: Read sequence

        Returns:
            Detection result with barcode, UMI, and feature positions
        """
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    def _find_barcode_umi_fwd(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        polyt_start = find_polyt_start(sequence)

        r1_start, r1_end = None, None
        if polyt_start != -1:
            # use relaxed parameters if polyA is found
            r1_occurrences = self.r1_indexer.get_occurrences(sequence[0:polyt_start + 1])
            r1_start, r1_end = detect_exact_positions(sequence, 0, polyt_start + 1,
                                                      self.r1_indexer.k, self.R1,
                                                      r1_occurrences, min_score=11,
                                                      end_delta=self.TERMINAL_MATCH_DELTA)

        if r1_start is None:
            # if polyT was not found, or linker was not found to the left of polyT,
            # look for linker in the entire read
            r1_occurrences = self.r1_indexer.get_occurrences(sequence)
            r1_start, r1_end = detect_exact_positions(sequence, 0, len(sequence),
                                                      self.r1_indexer.k, self.R1,
                                                      r1_occurrences, min_score=18,
                                                      start_delta=self.STRICT_TERMINAL_MATCH_DELTA,
                                                      end_delta=self.STRICT_TERMINAL_MATCH_DELTA)

        if r1_start is None:
            return TenXBarcodeDetectionResult(read_id, polyT=polyt_start)
        logger.debug("LINKER: %d-%d" % (r1_start, r1_end))

        if polyt_start == -1 or polyt_start - r1_end > self.BARCODE_LEN_10X + self.UMI_LEN + 10:
            # if polyT was not detected earlier, use relaxed parameters once the linker is found
            presumable_polyt_start = r1_end + self.BARCODE_LEN_10X + self.UMI_LEN
            search_start = presumable_polyt_start - 4
            search_end = min(len(sequence), presumable_polyt_start + 10)
            polyt_start = find_polyt_start(sequence[search_start:search_end], window_size=5, polya_fraction=1.0)
            if polyt_start != -1:
                polyt_start += search_start

        barcode_start = r1_end + 1
        barcode_end = r1_end + self.BARCODE_LEN_10X + 1
        potential_barcode = sequence[barcode_start:barcode_end + 1]
        logger.debug("Barcode: %s" % (potential_barcode))
        matching_barcodes = self.barcode_indexer.get_occurrences(potential_barcode,
                                                                 max_hits=self.max_barcodes_hits,
                                                                 min_kmers=self.min_matching_kmers)
        barcode, bc_score, bc_start, bc_end = \
            find_candidate_with_max_score_ssw(matching_barcodes, potential_barcode,
                                              min_score=self.min_score,
                                              score_diff=self.score_diff)

        if barcode is None:
            return TenXBarcodeDetectionResult(read_id, polyT=polyt_start, r1=r1_end)

        logger.debug("Found: %s %d-%d" % (barcode, bc_start, bc_end))
        # position of barcode end in the reference: end of potential barcode minus bases to the alignment end
        read_barcode_end = barcode_start + bc_end - 1
        potential_umi_start = read_barcode_end + 1
        potential_umi_end = polyt_start - 1
        if potential_umi_end - potential_umi_start <= 5:
            potential_umi_end = potential_umi_start + self.UMI_LEN - 1
        potential_umi = sequence[potential_umi_start:potential_umi_end + 1]
        logger.debug("Potential UMI: %s" % potential_umi)

        good_umi = False
        umi = potential_umi
        if self.UMI_LEN - self.UMI_LEN_DELTA <= len(umi) <= self.UMI_LEN + self.UMI_LEN_DELTA:
            good_umi = True

        if not umi:
            return TenXBarcodeDetectionResult(read_id, barcode, BC_score=bc_score, polyT=polyt_start, r1=r1_end)
        return TenXBarcodeDetectionResult(read_id, barcode, umi, bc_score, good_umi, polyT=polyt_start, r1=r1_end)

    def find_barcode_umi_no_polya(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        """Find barcode without polyA requirement."""
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    @staticmethod
    def result_type():
        return TenXBarcodeDetectionResult

    @classmethod
    def header(cls):
        return cls.result_type().header()


class TenXv2BarcodeDetector(TenXBarcodeDetector):
    """TenX v2 barcode detector."""
    UMI_LEN = 10

    def __init__(self, barcode_list: List[str]):
        """Initialize TenX v2 detector."""
        super().__init__(barcode_list)


class VisiumHDBarcodeDetector:
    """Visium HD spatial transcriptomics barcode detector."""

    R1 = "CTACACGACGCTCTTCCGATCT"  # 10x 3'
    BARCODE1_LEN_VIS = 16
    BARCODE2_LEN_VIS = 15
    SEPARATOR_BASES = 2
    TOTAL_BARCODE_LEN_VIS = BARCODE1_LEN_VIS + BARCODE2_LEN_VIS
    UMI_LEN_VIS = 9

    UMI_LEN_DELTA = 2
    TERMINAL_MATCH_DELTA = 2
    STRICT_TERMINAL_MATCH_DELTA = 1

    def __init__(self, barcode_pair_list: List[List[str]]):
        """
        Initialize Visium HD detector.

        Args:
            barcode_pair_list: List of two lists [part1_barcodes, part2_barcodes]
        """
        assert len(barcode_pair_list) == 2
        self.r1_indexer = KmerIndexer([VisiumHDBarcodeDetector.R1], kmer_size=7)
        self.part1_list = barcode_pair_list[0]
        self.part2_list = barcode_pair_list[1]

        self.part1_barcode_indexer = ArrayKmerIndexer(self.part1_list, kmer_size=7)
        self.part2_barcode_indexer = ArrayKmerIndexer(self.part2_list, kmer_size=7)
        self.max_barcodes_hits = 20
        self.min_matching_kmers = 2
        self.min_score = 13

        logger.debug("Min score set to %d" % self.min_score)

    def find_barcode_umi(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        """Detect barcode and UMI in a read sequence."""
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    def _find_barcode_umi_fwd(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        logger.debug("===== " + read_id)
        polyt_start = find_polyt_start(sequence)

        r1_start, r1_end = None, None
        if polyt_start != -1:
            # use relaxed parameters if polyA is found
            r1_occurrences = self.r1_indexer.get_occurrences(sequence[0:polyt_start + 1])
            r1_start, r1_end = detect_exact_positions(sequence, 0, polyt_start + 1,
                                                      self.r1_indexer.k, self.R1,
                                                      r1_occurrences, min_score=10,
                                                      end_delta=self.TERMINAL_MATCH_DELTA)

        if r1_start is None:
            # if polyT was not found, or linker was not found to the left of polyT,
            # look for linker in the entire read
            r1_occurrences = self.r1_indexer.get_occurrences(sequence)
            r1_start, r1_end = detect_exact_positions(sequence, 0, len(sequence),
                                                      self.r1_indexer.k, self.R1,
                                                      r1_occurrences, min_score=17,
                                                      start_delta=self.STRICT_TERMINAL_MATCH_DELTA,
                                                      end_delta=self.STRICT_TERMINAL_MATCH_DELTA)

        if r1_start is not None:
            logger.debug("PRIMER: %d-%d" % (r1_start, r1_end))

        if r1_end is not None and (polyt_start == -1 or polyt_start - r1_end > self.TOTAL_BARCODE_LEN_VIS + self.UMI_LEN_VIS + 10):
            # if polyT was not detected earlier, use relaxed parameters once the linker is found
            presumable_polyt_start = r1_end + self.TOTAL_BARCODE_LEN_VIS + self.UMI_LEN_VIS
            search_start = presumable_polyt_start - 4
            search_end = min(len(sequence), presumable_polyt_start + 10)
            polyt_start = find_polyt_start(sequence[search_start:search_end], window_size=5, polya_fraction=1.0)
            if polyt_start != -1:
                polyt_start += search_start

        if polyt_start == -1:
            if r1_start is None:
                return TenXBarcodeDetectionResult(read_id, polyT=polyt_start)
            # no polyT, start from the left
            potential_umi_start = r1_end + 1
            potential_umi_end = potential_umi_start + self.UMI_LEN_VIS - 1
            potential_umi = sequence[potential_umi_start:potential_umi_end + 1]
            logger.debug("Potential UMI: %s" % potential_umi)

            barcode1_start = r1_end + self.UMI_LEN_VIS + 1
            barcode1_end = barcode1_start + self.BARCODE1_LEN_VIS - 1
            potential_barcode1 = sequence[barcode1_start:barcode1_end + 1]
            matching_barcodes1 = self.part1_barcode_indexer.get_occurrences(potential_barcode1, max_hits=self.max_barcodes_hits)
            barcode1, bc1_score, bc1_start, bc1_end = \
                find_candidate_with_max_score_ssw_var_len(matching_barcodes1, potential_barcode1, min_score=self.min_score)
            logger.debug("Barcode 1: %s, %s" % (potential_barcode1, barcode1))
            real_bc1_end = barcode1_start + bc1_end

            barcode2_start = real_bc1_end + 1
            barcode2_end = barcode2_start + self.BARCODE2_LEN_VIS - 1
            potential_barcode2 = sequence[barcode2_start:barcode2_end + 1]
            matching_barcodes2 = self.part2_barcode_indexer.get_occurrences(potential_barcode2, max_hits=self.max_barcodes_hits)
            barcode2, bc2_score, bc2_start, bc2_end = \
                find_candidate_with_max_score_ssw_var_len(matching_barcodes2, potential_barcode2, min_score=self.min_score)
            logger.debug("Barcode 2: %s, %s" % (potential_barcode2, barcode2))

            if barcode1 is None or barcode2 is None:
                return TenXBarcodeDetectionResult(read_id, polyT=polyt_start, r1=r1_end)

            return TenXBarcodeDetectionResult(read_id, barcode1 + "|" + barcode2, potential_umi, bc1_score + bc2_score,
                                              UMI_good=True, polyT=polyt_start, r1=r1_end)

        barcode2_end = polyt_start - 1 - self.SEPARATOR_BASES
        barcode2_start = barcode2_end - self.BARCODE2_LEN_VIS + 1
        potential_barcode2 = sequence[barcode2_start:barcode2_end + 1]
        matching_barcodes2 = self.part2_barcode_indexer.get_occurrences(potential_barcode2, max_hits=self.max_barcodes_hits)
        barcode2, bc2_score, bc2_start, bc2_end = \
            find_candidate_with_max_score_ssw_var_len(matching_barcodes2, potential_barcode2, min_score=self.min_score)
        logger.debug("Barcode 2: %s, %s" % (potential_barcode2, barcode2))

        real_bc2_start = barcode2_start + bc2_start
        barcode1_end = real_bc2_start - 1
        barcode1_start = barcode1_end - self.BARCODE1_LEN_VIS + 1
        potential_barcode1 = sequence[barcode1_start:barcode1_end + 1]
        matching_barcodes1 = self.part1_barcode_indexer.get_occurrences(potential_barcode1, max_hits=self.max_barcodes_hits)
        barcode1, bc1_score, bc1_start, bc1_end = \
            find_candidate_with_max_score_ssw_var_len(matching_barcodes1, potential_barcode1, min_score=self.min_score)
        logger.debug("Barcode 1: %s, %s" % (potential_barcode1, barcode1))
        real_bc1_start = barcode1_start + bc1_start

        potential_umi_end = real_bc1_start - 1
        if r1_end is not None:
            potential_umi_start = r1_end + 1
        else:
            potential_umi_start = max(0, potential_umi_end - self.UMI_LEN_VIS)
        umi_good = abs(potential_umi_end - potential_umi_start + 1 - self.UMI_LEN_VIS) <= self.UMI_LEN_DELTA
        potential_umi = sequence[potential_umi_start:potential_umi_end + 1]
        logger.debug("Potential UMI: %s" % potential_umi)

        if barcode1 is None or barcode2 is None:
            return TenXBarcodeDetectionResult(read_id, polyT=polyt_start, r1=r1_end if r1_end is not None else -1)

        return TenXBarcodeDetectionResult(read_id, barcode1 + "|" + barcode2, potential_umi, bc1_score + bc2_score,
                                          UMI_good=umi_good, polyT=polyt_start, r1=r1_end if r1_end is not None else -1)

    def find_barcode_umi_no_polya(self, read_id: str, sequence: str) -> TenXBarcodeDetectionResult:
        """Find barcode without polyA requirement."""
        read_result = self._find_barcode_umi_fwd(read_id, sequence)
        if read_result.polyT != -1:
            read_result.set_strand("+")
        if read_result.is_valid():
            return read_result

        rev_seq = reverese_complement(sequence)
        read_rev_result = self._find_barcode_umi_fwd(read_id, rev_seq)
        if read_rev_result.polyT != -1:
            read_rev_result.set_strand("-")
        if read_rev_result.is_valid():
            return read_rev_result

        return read_result if read_result.more_informative_than(read_rev_result) else read_rev_result

    @staticmethod
    def result_type():
        return TenXBarcodeDetectionResult

    @classmethod
    def header(cls):
        return cls.result_type().header()
